# -*- coding=utf-8 -*-
# Copyright 2017 The Abseil Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for absl.command_name."""

import ctypes
import errno
import os
import unittest
from unittest import mock

from absl import command_name
from absl.testing import absltest


def _get_kernel_process_name():
  """Returns the Kernel's name for our process or an empty string."""
  try:
    with open('/proc/self/status', 'rt') as status_file:
      for line in status_file:
        if line.startswith('Name:'):
          return line.split(':', 2)[1].strip().encode('ascii', 'replace')
      return b''
  except IOError:
    return b''


def _is_prctl_syscall_available():
  try:
    libc = ctypes.CDLL('libc.so.6', use_errno=True)
  except OSError:
    return False
  zero = ctypes.c_ulong(0)
  try:
    status = libc.prctl(zero, zero, zero, zero, zero)
  except AttributeError:
    return False
  if status < 0 and errno.ENOSYS == ctypes.get_errno():
    return False
  return True


@unittest.skipIf(not _get_kernel_process_name(),
                 '_get_kernel_process_name() fails.')
class CommandNameTest(absltest.TestCase):

  def assertProcessNameSimilarTo(self, new_name):
    if not isinstance(new_name, bytes):
      new_name = new_name.encode('ascii', 'replace')
    actual_name = _get_kernel_process_name()
    self.assertTrue(actual_name)
    self.assertTrue(new_name.startswith(actual_name),
                    msg='set {!r} vs found {!r}'.format(new_name, actual_name))

  @unittest.skipIf(not os.access('/proc/self/comm', os.W_OK),
                   '/proc/self/comm is not writeable.')
  def test_set_kernel_process_name(self):
    new_name = u'ProcessNam0123456789abcdefghijklmnöp'
    command_name.set_kernel_process_name(new_name)
    self.assertProcessNameSimilarTo(new_name)

  @unittest.skipIf(not _is_prctl_syscall_available(),
                   'prctl() system call missing from libc.so.6.')
  def test_set_kernel_process_name_no_proc_file(self):
    new_name = b'NoProcFile0123456789abcdefghijklmnop'
    mock_open = mock.mock_open()
    with mock.patch.object(command_name, 'open', mock_open, create=True):
      mock_open.side_effect = IOError('mock open that raises.')
      command_name.set_kernel_process_name(new_name)
    mock_open.assert_called_with('/proc/self/comm', mock.ANY)
    self.assertProcessNameSimilarTo(new_name)

  def test_set_kernel_process_name_failure(self):
    starting_name = _get_kernel_process_name()
    new_name = b'NameTest'
    mock_open = mock.mock_open()
    mock_ctypes_cdll = mock.patch('ctypes.CDLL')
    with mock.patch.object(command_name, 'open', mock_open, create=True):
      with mock.patch('ctypes.CDLL') as mock_ctypes_cdll:
        mock_open.side_effect = IOError('mock open that raises.')
        mock_libc = mock.Mock(['prctl'])
        mock_ctypes_cdll.return_value = mock_libc
        command_name.set_kernel_process_name(new_name)
    mock_open.assert_called_with('/proc/self/comm', mock.ANY)
    self.assertEqual(1, mock_libc.prctl.call_count)
    self.assertEqual(starting_name, _get_kernel_process_name())  # No change.

  def test_make_process_name_useful(self):
    test_name = 'hello.from.test'
    with mock.patch('sys.argv', [test_name]):
      command_name.make_process_name_useful()
    self.assertProcessNameSimilarTo(test_name)


if __name__ == '__main__':
  absltest.main()



# Copyright 2017 The Abseil Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Additional tests for flag argument parsers.

Most of the argument parsers are covered in the flags_test.py.
"""

import enum

from absl.flags import _argument_parser
from absl.testing import absltest
from absl.testing import parameterized


class ArgumentParserTest(absltest.TestCase):

  def test_instance_cache(self):
    parser1 = _argument_parser.FloatParser()
    parser2 = _argument_parser.FloatParser()
    self.assertIs(parser1, parser2)

  def test_parse_wrong_type(self):
    parser = _argument_parser.ArgumentParser()
    with self.assertRaises(TypeError):
      parser.parse(0)  # type: ignore

    if bytes is not str:
      # In PY3, it does not accept bytes.
      with self.assertRaises(TypeError):
        parser.parse(b'')  # type: ignore


class BooleanParserTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self.parser = _argument_parser.BooleanParser()

  def test_parse_bytes(self):
    with self.assertRaises(TypeError):
      self.parser.parse(b'true')  # type: ignore

  def test_parse_str(self):
    self.assertTrue(self.parser.parse('true'))

  def test_parse_unicode(self):
    self.assertTrue(self.parser.parse(u'true'))

  def test_parse_wrong_type(self):
    with self.assertRaises(TypeError):
      self.parser.parse(1.234)  # type: ignore

  def test_parse_str_false(self):
    self.assertFalse(self.parser.parse('false'))

  def test_parse_integer(self):
    self.assertTrue(self.parser.parse(1))

  def test_parse_invalid_integer(self):
    with self.assertRaises(ValueError):
      self.parser.parse(-1)

  def test_parse_invalid_str(self):
    with self.assertRaises(ValueError):
      self.parser.parse('nottrue')


class FloatParserTest(absltest.TestCase):

  def setUp(self):
    self.parser = _argument_parser.FloatParser()

  def test_parse_string(self):
    self.assertEqual(1.5, self.parser.parse('1.5'))

  def test_parse_wrong_type(self):
    with self.assertRaises(TypeError):
      self.parser.parse(False)  # type: ignore


class IntegerParserTest(absltest.TestCase):

  def setUp(self):
    self.parser = _argument_parser.IntegerParser()

  def test_parse_string(self):
    self.assertEqual(1, self.parser.parse('1'))

  def test_parse_wrong_type(self):
    with self.assertRaises(TypeError):
      self.parser.parse(1e2)  # type: ignore
    with self.assertRaises(TypeError):
      self.parser.parse(False)  # type: ignore


class EnumParserTest(absltest.TestCase):

  def test_empty_values(self):
    with self.assertRaises(ValueError):
      _argument_parser.EnumParser([])

  def test_parse(self):
    parser = _argument_parser.EnumParser(['apple', 'banana'])
    self.assertEqual('apple', parser.parse('apple'))

  def test_parse_not_found(self):
    parser = _argument_parser.EnumParser(['apple', 'banana'])
    with self.assertRaises(ValueError):
      parser.parse('orange')


class Fruit(enum.Enum):
  APPLE = 1
  BANANA = 2


class EmptyEnum(enum.Enum):
  pass


class MixedCaseEnum(enum.Enum):
  APPLE = 1
  BANANA = 2
  apple = 3


class EnumClassParserTest(parameterized.TestCase):

  def test_requires_enum(self):
    with self.assertRaises(TypeError):
      _argument_parser.EnumClassParser(['apple', 'banana'])  # type: ignore

  def test_requires_non_empty_enum_class(self):
    with self.assertRaises(ValueError):
      _argument_parser.EnumClassParser(EmptyEnum)

  def test_case_sensitive_rejects_duplicates(self):
    unused_normal_parser = _argument_parser.EnumClassParser(MixedCaseEnum)
    with self.assertRaisesRegex(ValueError, 'Duplicate.+apple'):
      _argument_parser.EnumClassParser(MixedCaseEnum, case_sensitive=False)

  def test_parse_string(self):
    parser = _argument_parser.EnumClassParser(Fruit)
    self.assertEqual(Fruit.APPLE, parser.parse('APPLE'))

  def test_parse_string_case_sensitive(self):
    parser = _argument_parser.EnumClassParser(Fruit)
    with self.assertRaises(ValueError):
      parser.parse('apple')

  @parameterized.parameters('APPLE', 'apple', 'Apple')
  def test_parse_string_case_insensitive(self, value):
    parser = _argument_parser.EnumClassParser(Fruit, case_sensitive=False)
    self.assertIs(Fruit.APPLE, parser.parse(value))

  def test_parse_literal(self):
    parser = _argument_parser.EnumClassParser(Fruit)
    self.assertEqual(Fruit.APPLE, parser.parse(Fruit.APPLE))

  def test_parse_not_found(self):
    parser = _argument_parser.EnumClassParser(Fruit)
    with self.assertRaises(ValueError):
      parser.parse('ORANGE')

  @parameterized.parameters((Fruit.BANANA, False, 'BANANA'),
                            (Fruit.BANANA, True, 'banana'))
  def test_serialize_parse(self, value, lowercase, expected):
    serializer = _argument_parser.EnumClassSerializer(lowercase=lowercase)
    parser = _argument_parser.EnumClassParser(
        Fruit, case_sensitive=not lowercase)
    serialized = serializer.serialize(value)
    self.assertEqual(serialized, expected)
    self.assertEqual(value, parser.parse(expected))


class SerializerTest(parameterized.TestCase):

  def test_csv_serializer(self):
    serializer = _argument_parser.CsvListSerializer('+')
    self.assertEqual(serializer.serialize(['foo', 'bar']), 'foo+bar')

  @parameterized.parameters([
      dict(lowercase=False, expected='APPLE+BANANA'),
      dict(lowercase=True, expected='apple+banana'),
  ])
  def test_enum_class_list_serializer(self, lowercase, expected):
    values = [Fruit.APPLE, Fruit.BANANA]
    serializer = _argument_parser.EnumClassListSerializer(
        list_sep='+', lowercase=lowercase)
    serialized = serializer.serialize(values)
    self.assertEqual(expected, serialized)


class HelperFunctionsTest(absltest.TestCase):

  def test_is_integer_type(self):
    self.assertTrue(_argument_parser._is_integer_type(1))
    # Note that isinstance(False, int) == True.
    self.assertFalse(_argument_parser._is_integer_type(False))


if __name__ == '__main__':
  absltest.main()



# Copyright 2018 The Abseil Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test helper for argparse_flags_test."""

import os
import random

from absl import app
from absl import flags
from absl.flags import argparse_flags

FLAGS = flags.FLAGS

flags.DEFINE_string('absl_echo', None, 'The echo message from absl.flags.')


def parse_flags_simple(argv):
  """Simple example for absl.flags + argparse."""
  parser = argparse_flags.ArgumentParser(
      description='A simple example of argparse_flags.')
  parser.add_argument(
      '--argparse_echo', help='The echo message from argparse_flags')
  return parser.parse_args(argv[1:])


def main_simple(args):
  print('--absl_echo is', FLAGS.absl_echo)
  print('--argparse_echo is', args.argparse_echo)


def roll_dice(args):
  print('Rolled a dice:', random.randint(1, args.num_faces))


def shuffle(args):
  inputs = list(args.inputs)
  random.shuffle(inputs)
  print('Shuffled:', ' '.join(inputs))


def parse_flags_subcommands(argv):
  """Subcommands example for absl.flags + argparse."""
  parser = argparse_flags.ArgumentParser(
      description='A subcommands example of argparse_flags.')
  parser.add_argument('--argparse_echo',
                      help='The echo message from argparse_flags')

  subparsers = parser.add_subparsers(help='The command to execute.')

  roll_dice_parser = subparsers.add_parser(
      'roll_dice', help='Roll a dice.')
  roll_dice_parser.add_argument('--num_faces', type=int, default=6)
  roll_dice_parser.set_defaults(command=roll_dice)

  shuffle_parser = subparsers.add_parser(
      'shuffle', help='Shuffle inputs.')
  shuffle_parser.add_argument(
      'inputs', metavar='I', nargs='+', help='Inputs to shuffle.')
  shuffle_parser.set_defaults(command=shuffle)

  return parser.parse_args(argv[1:])


def main_subcommands(args):
  main_simple(args)
  args.command(args)


if __name__ == '__main__':
  main_func_name = os.environ['MAIN_FUNC']
  flags_parser_func_name = os.environ['FLAGS_PARSER_FUNC']
  app.run(main=globals()[main_func_name],
          flags_parser=globals()[flags_parser_func_name])



# Copyright 2017 The Abseil Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Internal helper functions for Abseil Python flags library."""

import os
import re
import struct
import sys
import textwrap
import types
from typing import Any, Dict, Iterable, List, NamedTuple, Optional, Sequence, Set
from xml.dom import minidom
# pylint: disable=g-import-not-at-top
try:
  import fcntl
except ImportError:
  fcntl = None
try:
  # Importing termios will fail on non-unix platforms.
  import termios
except ImportError:
  termios = None
# pylint: enable=g-import-not-at-top


_DEFAULT_HELP_WIDTH = 80  # Default width of help output.
# Minimal "sane" width of help output. We assume that any value below 40 is
# unreasonable.
_MIN_HELP_WIDTH = 40

# Define the allowed error rate in an input string to get suggestions.
#
# We lean towards a high threshold because we tend to be matching a phrase,
# and the simple algorithm used here is geared towards correcting word
# spellings.
#
# For manual testing, consider "<command> --list" which produced a large number
# of spurious suggestions when we used "least_errors > 0.5" instead of
# "least_erros >= 0.5".
_SUGGESTION_ERROR_RATE_THRESHOLD = 0.50

# Characters that cannot appear or are highly discouraged in an XML 1.0
# document. (See http://www.w3.org/TR/REC-xml/#charsets or
# https://en.wikipedia.org/wiki/Valid_characters_in_XML#XML_1.0)
_ILLEGAL_XML_CHARS_REGEX = re.compile(
    u'[\x00-\x08\x0b\x0c\x0e-\x1f\x7f-\x84\x86-\x9f\ud800-\udfff\ufffe\uffff]')

# This is a set of module ids for the modules that disclaim key flags.
# This module is explicitly added to this set so that we never consider it to
# define key flag.
disclaim_module_ids: Set[int] = set([id(sys.modules[__name__])])


# Define special flags here so that help may be generated for them.
# NOTE: Please do NOT use SPECIAL_FLAGS from outside flags module.
# Initialized inside flagvalues.py.
# NOTE: This cannot be annotated as its actual FlagValues type since this would
# create a circular dependency.
SPECIAL_FLAGS: Any = None


# This points to the flags module, initialized in flags/__init__.py.
# This should only be used in adopt_module_key_flags to take SPECIAL_FLAGS into
# account.
FLAGS_MODULE: types.ModuleType = None


class _ModuleObjectAndName(NamedTuple):
  """Module object and name.

  Fields:
  - module: object, module object.
  - module_name: str, module name.
  """
  module: types.ModuleType
  module_name: str


def get_module_object_and_name(
    globals_dict: Dict[str, Any]
) -> _ModuleObjectAndName:
  """Returns the module that defines a global environment, and its name.

  Args:
    globals_dict: A dictionary that should correspond to an environment
      providing the values of the globals.

  Returns:
    _ModuleObjectAndName - pair of module object & module name.
    Returns (None, None) if the module could not be identified.
  """
  name = globals_dict.get('__name__', None)
  module = sys.modules.get(name, None)
  # Pick a more informative name for the main module.
  return _ModuleObjectAndName(module,
                              (sys.argv[0] if name == '__main__' else name))


def get_calling_module_object_and_name() -> _ModuleObjectAndName:
  """Returns the module that's calling into this module.

  We generally use this function to get the name of the module calling a
  DEFINE_foo... function.

  Returns:
    The module object that called into this one.

  Raises:
    AssertionError: Raised when no calling module could be identified.
  """
  for depth in range(1, sys.getrecursionlimit()):
    # sys._getframe is the right thing to use here, as it's the best
    # way to walk up the call stack.
    globals_for_frame = sys._getframe(depth).f_globals  # pylint: disable=protected-access
    module, module_name = get_module_object_and_name(globals_for_frame)
    if id(module) not in disclaim_module_ids and module_name is not None:
      return _ModuleObjectAndName(module, module_name)
  raise AssertionError('No module was found')


def get_calling_module() -> str:
  """Returns the name of the module that's calling into this module."""
  return get_calling_module_object_and_name().module_name


def create_xml_dom_element(
    doc: minidom.Document, name: str, value: Any
) -> minidom.Element:
  """Returns an XML DOM element with name and text value.

  Args:
    doc: minidom.Document, the DOM document it should create nodes from.
    name: str, the tag of XML element.
    value: object, whose string representation will be used
        as the value of the XML element. Illegal or highly discouraged xml 1.0
        characters are stripped.

  Returns:
    An instance of minidom.Element.
  """
  s = str(value)
  if isinstance(value, bool):
    # Display boolean values as the C++ flag library does: no caps.
    s = s.lower()
  # Remove illegal xml characters.
  s = _ILLEGAL_XML_CHARS_REGEX.sub(u'', s)

  e = doc.createElement(name)
  e.appendChild(doc.createTextNode(s))
  return e


def get_help_width() -> int:
  """Returns the integer width of help lines that is used in TextWrap."""
  if not sys.stdout.isatty() or termios is None or fcntl is None:
    return _DEFAULT_HELP_WIDTH
  try:
    data = fcntl.ioctl(sys.stdout, termios.TIOCGWINSZ, b'1234')
    columns = struct.unpack('hh', data)[1]
    # Emacs mode returns 0.
    # Here we assume that any value below 40 is unreasonable.
    if columns >= _MIN_HELP_WIDTH:
      return columns
    # Returning an int as default is fine, int(int) just return the int.
    return int(os.getenv('COLUMNS', _DEFAULT_HELP_WIDTH))

  except (TypeError, IOError, struct.error):
    return _DEFAULT_HELP_WIDTH


def get_flag_suggestions(
    attempt: Optional[str], longopt_list: Sequence[str]
) -> List[str]:
  """Returns helpful similar matches for an invalid flag."""
  # Don't suggest on very short strings, or if no longopts are specified.
  if len(attempt) <= 2 or not longopt_list:
    return []

  option_names = [v.split('=')[0] for v in longopt_list]

  # Find close approximations in flag prefixes.
  # This also handles the case where the flag is spelled right but ambiguous.
  distances = [(_damerau_levenshtein(attempt, option[0:len(attempt)]), option)
               for option in option_names]
  # t[0] is distance, and sorting by t[1] allows us to have stable output.
  distances.sort()

  least_errors, _ = distances[0]
  # Don't suggest excessively bad matches.
  if least_errors >= _SUGGESTION_ERROR_RATE_THRESHOLD * len(attempt):
    return []

  suggestions = []
  for errors, name in distances:
    if errors == least_errors:
      suggestions.append(name)
    else:
      break
  return suggestions


def _damerau_levenshtein(a, b):
  """Returns Damerau-Levenshtein edit distance from a to b."""
  memo = {}

  def distance(x, y):
    """Recursively defined string distance with memoization."""
    if (x, y) in memo:
      return memo[x, y]
    if not x:
      d = len(y)
    elif not y:
      d = len(x)
    else:
      d = min(
          distance(x[1:], y) + 1,  # correct an insertion error
          distance(x, y[1:]) + 1,  # correct a deletion error
          distance(x[1:], y[1:]) + (x[0] != y[0]))  # correct a wrong character
      if len(x) >= 2 and len(y) >= 2 and x[0] == y[1] and x[1] == y[0]:
        # Correct a transposition.
        t = distance(x[2:], y[2:]) + 1
        if d > t:
          d = t

    memo[x, y] = d
    return d
  return distance(a, b)


def text_wrap(
    text: str,
    length: Optional[int] = None,
    indent: str = '',
    firstline_indent: Optional[str] = None,
) -> str:
  """Wraps a given text to a maximum line length and returns it.

  It turns lines that only contain whitespace into empty lines, keeps new lines,
  and expands tabs using 4 spaces.

  Args:
    text: str, text to wrap.
    length: int, maximum length of a line, includes indentation.
        If this is None then use get_help_width()
    indent: str, indent for all but first line.
    firstline_indent: str, indent for first line; if None, fall back to indent.

  Returns:
    str, the wrapped text.

  Raises:
    ValueError: Raised if indent or firstline_indent not shorter than length.
  """
  # Get defaults where callee used None
  if length is None:
    length = get_help_width()
  if indent is None:
    indent = ''
  if firstline_indent is None:
    firstline_indent = indent

  if len(indent) >= length:
    raise ValueError('Length of indent exceeds length')
  if len(firstline_indent) >= length:
    raise ValueError('Length of first line indent exceeds length')

  text = text.expandtabs(4)

  result = []
  # Create one wrapper for the first paragraph and one for subsequent
  # paragraphs that does not have the initial wrapping.
  wrapper = textwrap.TextWrapper(
      width=length, initial_indent=firstline_indent, subsequent_indent=indent)
  subsequent_wrapper = textwrap.TextWrapper(
      width=length, initial_indent=indent, subsequent_indent=indent)

  # textwrap does not have any special treatment for newlines. From the docs:
  # "...newlines may appear in the middle of a line and cause strange output.
  # For this reason, text should be split into paragraphs (using
  # str.splitlines() or similar) which are wrapped separately."
  for paragraph in (p.strip() for p in text.splitlines()):
    if paragraph:
      result.extend(wrapper.wrap(paragraph))
    else:
      result.append('')  # Keep empty lines.
    # Replace initial wrapper with wrapper for subsequent paragraphs.
    wrapper = subsequent_wrapper

  return '\n'.join(result)


def flag_dict_to_args(
    flag_map: Dict[str, Any], multi_flags: Optional[Set[str]] = None
) -> Iterable[str]:
  """Convert a dict of values into process call parameters.

  This method is used to convert a dictionary into a sequence of parameters
  for a binary that parses arguments using this module.

  Args:
    flag_map: dict, a mapping where the keys are flag names (strings).
        values are treated according to their type:

        * If value is ``None``, then only the name is emitted.
        * If value is ``True``, then only the name is emitted.
        * If value is ``False``, then only the name prepended with 'no' is
          emitted.
        * If value is a string then ``--name=value`` is emitted.
        * If value is a collection, this will emit
          ``--name=value1,value2,value3``, unless the flag name is in
          ``multi_flags``, in which case this will emit
          ``--name=value1 --name=value2 --name=value3``.
        * Everything else is converted to string an passed as such.

    multi_flags: set, names (strings) of flags that should be treated as
        multi-flags.
  Yields:
    sequence of string suitable for a subprocess execution.
  """
  for key, value in flag_map.items():
    if value is None:
      yield '--%s' % key
    elif isinstance(value, bool):
      if value:
        yield '--%s' % key
      else:
        yield '--no%s' % key
    elif isinstance(value, (bytes, type(u''))):
      # We don't want strings to be handled like python collections.
      yield '--%s=%s' % (key, value)
    else:
      # Now we attempt to deal with collections.
      try:
        if multi_flags and key in multi_flags:
          for item in value:
            yield '--%s=%s' % (key, str(item))
        else:
          yield '--%s=%s' % (key, ','.join(str(item) for item in value))
      except TypeError:
        # Default case.
        yield '--%s=%s' % (key, value)


def trim_docstring(docstring: str) -> str:
  """Removes indentation from triple-quoted strings.

  This is the function specified in PEP 257 to handle docstrings:
  https://www.python.org/dev/peps/pep-0257/.

  Args:
    docstring: str, a python docstring.

  Returns:
    str, docstring with indentation removed.
  """
  if not docstring:
    return ''

  # If you've got a line longer than this you have other problems...
  max_indent = 1 << 29

  # Convert tabs to spaces (following the normal Python rules)
  # and split into a list of lines:
  lines = docstring.expandtabs().splitlines()

  # Determine minimum indentation (first line doesn't count):
  indent = max_indent
  for line in lines[1:]:
    stripped = line.lstrip()
    if stripped:
      indent = min(indent, len(line) - len(stripped))
  # Remove indentation (first line is special):
  trimmed = [lines[0].strip()]
  if indent < max_indent:
    for line in lines[1:]:
      trimmed.append(line[indent:].rstrip())
  # Strip off trailing and leading blank lines:
  while trimmed and not trimmed[-1]:
    trimmed.pop()
  while trimmed and not trimmed[0]:
    trimmed.pop(0)
  # Return a single string:
  return '\n'.join(trimmed)


def doc_to_help(doc: str) -> str:
  """Takes a __doc__ string and reformats it as help."""

  # Get rid of starting and ending white space. Using lstrip() or even
  # strip() could drop more than maximum of first line and right space
  # of last line.
  doc = doc.strip()

  # Get rid of all empty lines.
  whitespace_only_line = re.compile('^[ \t]+$', re.M)
  doc = whitespace_only_line.sub('', doc)

  # Cut out common space at line beginnings.
  doc = trim_docstring(doc)

  # Just like this module's comment, comments tend to be aligned somehow.
  # In other words they all start with the same amount of white space.
  # 1) keep double new lines;
  # 2) keep ws after new lines if not empty line;
  # 3) all other new lines shall be changed to a space;
  # Solution: Match new lines between non white space and replace with space.
  doc = re.sub(r'(?<=\S)\n(?=\S)', ' ', doc, flags=re.M)

  return doc



# Copyright 2017 The Abseil Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A Python test reporter that generates test reports in JUnit XML format."""

import datetime
import re
import sys
import threading
import time
import traceback
import unittest
from xml.sax import saxutils
from absl.testing import _pretty_print_reporter


# See http://www.w3.org/TR/REC-xml/#NT-Char
_bad_control_character_codes = set(range(0, 0x20)) - {0x9, 0xA, 0xD}


_control_character_conversions = {
    chr(i): '\\x{:02x}'.format(i) for i in _bad_control_character_codes}


_escape_xml_attr_conversions = {
    '"': '&quot;',
    "'": '&apos;',
    '\n': '&#xA;',
    '\t': '&#x9;',
    '\r': '&#xD;',
    ' ': '&#x20;'}
_escape_xml_attr_conversions.update(_control_character_conversions)


# When class or module level function fails, unittest/suite.py adds a
# _ErrorHolder instance instead of a real TestCase, and it has a description
# like "setUpClass (__main__.MyTestCase)".
_CLASS_OR_MODULE_LEVEL_TEST_DESC_REGEX = re.compile(r'^(\w+) \((\S+)\)$')


# NOTE: while saxutils.quoteattr() theoretically does the same thing; it
# seems to often end up being too smart for it's own good not escaping properly.
# This function is much more reliable.
def _escape_xml_attr(content):
  """Escapes xml attributes."""
  # Note: saxutils doesn't escape the quotes.
  return saxutils.escape(content, _escape_xml_attr_conversions)


def _escape_cdata(s):
  """Escapes a string to be used as XML CDATA.

  CDATA characters are treated strictly as character data, not as XML markup,
  but there are still certain restrictions on them.

  Args:
    s: the string to be escaped.
  Returns:
    An escaped version of the input string.
  """
  for char, escaped in _control_character_conversions.items():
    s = s.replace(char, escaped)
  return s.replace(']]>', ']] >')


def _iso8601_timestamp(timestamp):
  """Produces an ISO8601 datetime.

  Args:
    timestamp: an Epoch based timestamp in seconds.

  Returns:
    A iso8601 format timestamp if the input is a valid timestamp, None otherwise
  """
  if timestamp is None or timestamp < 0:
    return None
  return datetime.datetime.fromtimestamp(
      timestamp, tz=datetime.timezone.utc).isoformat()


def _print_xml_element_header(element, attributes, stream, indentation=''):
  """Prints an XML header of an arbitrary element.

  Args:
    element: element name (testsuites, testsuite, testcase)
    attributes: 2-tuple list with (attributes, values) already escaped
    stream: output stream to write test report XML to
    indentation: indentation added to the element header
  """
  stream.write('%s<%s' % (indentation, element))
  for attribute in attributes:
    if (len(attribute) == 2 and attribute[0] is not None and
        attribute[1] is not None):
      stream.write(' %s="%s"' % (attribute[0], attribute[1]))
  stream.write('>\n')

# Copy time.time which ensures the real time is used internally.
# This prevents bad interactions with tests that stub out time.
_time_copy = time.time

if hasattr(traceback, '_some_str'):
  # Use the traceback module str function to format safely.
  _safe_str = traceback._some_str
else:
  _safe_str = str  # pylint: disable=invalid-name


class _TestCaseResult(object):
  """Private helper for _TextAndXMLTestResult that represents a test result.

  Attributes:
    test: A TestCase instance of an individual test method.
    name: The name of the individual test method.
    full_class_name: The full name of the test class.
    run_time: The duration (in seconds) it took to run the test.
    start_time: Epoch relative timestamp of when test started (in seconds)
    errors: A list of error 4-tuples. Error tuple entries are
        1) a string identifier of either "failure" or "error"
        2) an exception_type
        3) an exception_message
        4) a string version of a sys.exc_info()-style tuple of values
           ('error', err[0], err[1], self._exc_info_to_string(err))
           If the length of errors is 0, then the test is either passed or
           skipped.
    skip_reason: A string explaining why the test was skipped.
  """

  def __init__(self, test):
    self.run_time = -1
    self.start_time = -1
    self.skip_reason = None
    self.errors = []
    self.test = test

    # Parse the test id to get its test name and full class path.
    # Unfortunately there is no better way of knowning the test and class.
    # Worse, unittest uses _ErrorHandler instances to represent class / module
    # level failures.
    test_desc = test.id() or str(test)
    # Check if it's something like "setUpClass (__main__.TestCase)".
    match = _CLASS_OR_MODULE_LEVEL_TEST_DESC_REGEX.match(test_desc)
    if match:
      name = match.group(1)
      full_class_name = match.group(2)
    else:
      class_name = unittest.util.strclass(test.__class__)
      if isinstance(test, unittest.case._SubTest):
        # If the test case is a _SubTest, the real TestCase instance is
        # available as _SubTest.test_case.
        class_name = unittest.util.strclass(test.test_case.__class__)
      if test_desc.startswith(class_name + '.'):
        # In a typical unittest.TestCase scenario, test.id() returns with
        # a class name formatted using unittest.util.strclass.
        name = test_desc[len(class_name)+1:]
        full_class_name = class_name
      else:
        # Otherwise make a best effort to guess the test name and full class
        # path.
        parts = test_desc.rsplit('.', 1)
        name = parts[-1]
        full_class_name = parts[0] if len(parts) == 2 else ''
    self.name = _escape_xml_attr(name)
    self.full_class_name = _escape_xml_attr(full_class_name)

  def set_run_time(self, time_in_secs):
    self.run_time = time_in_secs

  def set_start_time(self, time_in_secs):
    self.start_time = time_in_secs

  def print_xml_summary(self, stream):
    """Prints an XML Summary of a TestCase.

    Status and result are populated as per JUnit XML test result reporter.
    A test that has been skipped will always have a skip reason,
    as every skip method in Python's unittest requires the reason arg to be
    passed.

    Args:
      stream: output stream to write test report XML to
    """

    if self.skip_reason is None:
      status = 'run'
      result = 'completed'
    else:
      status = 'notrun'
      result = 'suppressed'

    test_case_attributes = [
        ('name', '%s' % self.name),
        ('status', '%s' % status),
        ('result', '%s' % result),
        ('time', '%.3f' % self.run_time),
        ('classname', self.full_class_name),
        ('timestamp', _iso8601_timestamp(self.start_time)),
    ]
    _print_xml_element_header('testcase', test_case_attributes, stream, '  ')
    self._print_testcase_details(stream)
    stream.write('  </testcase>\n')

  def _print_testcase_details(self, stream):
    for error in self.errors:
      outcome, exception_type, message, error_msg = error  # pylint: disable=unpacking-non-sequence
      message = _escape_xml_attr(_safe_str(message))
      exception_type = _escape_xml_attr(str(exception_type))
      error_msg = _escape_cdata(error_msg)
      stream.write('  <%s message="%s" type="%s"><![CDATA[%s]]></%s>\n'
                   % (outcome, message, exception_type, error_msg, outcome))


class _TestSuiteResult(object):
  """Private helper for _TextAndXMLTestResult."""

  def __init__(self):
    self.suites = {}
    self.failure_counts = {}
    self.error_counts = {}
    self.overall_start_time = -1
    self.overall_end_time = -1
    self._testsuites_properties = {}

  def add_test_case_result(self, test_case_result):
    suite_name = type(test_case_result.test).__name__
    if suite_name == '_ErrorHolder':
      # _ErrorHolder is a special case created by unittest for class / module
      # level functions.
      suite_name = test_case_result.full_class_name.rsplit('.')[-1]
    if isinstance(test_case_result.test, unittest.case._SubTest):
      # If the test case is a _SubTest, the real TestCase instance is
      # available as _SubTest.test_case.
      suite_name = type(test_case_result.test.test_case).__name__

    self._setup_test_suite(suite_name)
    self.suites[suite_name].append(test_case_result)
    for error in test_case_result.errors:
      # Only count the first failure or error so that the sum is equal to the
      # total number of *testcases* that have failures or errors.
      if error[0] == 'failure':
        self.failure_counts[suite_name] += 1
        break
      elif error[0] == 'error':
        self.error_counts[suite_name] += 1
        break

  def print_xml_summary(self, stream):
    overall_test_count = sum(len(x) for x in self.suites.values())
    overall_failures = sum(self.failure_counts.values())
    overall_errors = sum(self.error_counts.values())
    overall_attributes = [
        ('name', ''),
        ('tests', '%d' % overall_test_count),
        ('failures', '%d' % overall_failures),
        ('errors', '%d' % overall_errors),
        ('time', '%.3f' % (self.overall_end_time - self.overall_start_time)),
        ('timestamp', _iso8601_timestamp(self.overall_start_time)),
    ]
    _print_xml_element_header('testsuites', overall_attributes, stream)
    if self._testsuites_properties:
      stream.write('    <properties>\n')
      for name, value in sorted(self._testsuites_properties.items()):
        stream.write('      <property name="%s" value="%s"></property>\n' %
                     (_escape_xml_attr(name), _escape_xml_attr(str(value))))
      stream.write('    </properties>\n')

    for suite_name in self.suites:
      suite = self.suites[suite_name]
      suite_end_time = max(x.start_time + x.run_time for x in suite)
      suite_start_time = min(x.start_time for x in suite)
      failures = self.failure_counts[suite_name]
      errors = self.error_counts[suite_name]
      suite_attributes = [
          ('name', '%s' % suite_name),
          ('tests', '%d' % len(suite)),
          ('failures', '%d' % failures),
          ('errors', '%d' % errors),
          ('time', '%.3f' % (suite_end_time - suite_start_time)),
          ('timestamp', _iso8601_timestamp(suite_start_time)),
      ]
      _print_xml_element_header('testsuite', suite_attributes, stream)

      # test_case_result entries are not guaranteed to be in any user-friendly
      # order, especially when using subtests. So sort them.
      for test_case_result in sorted(suite, key=lambda t: t.name):
        test_case_result.print_xml_summary(stream)
      stream.write('</testsuite>\n')
    stream.write('</testsuites>\n')

  def _setup_test_suite(self, suite_name):
    """Adds a test suite to the set of suites tracked by this test run.

    Args:
      suite_name: string, The name of the test suite being initialized.
    """
    if suite_name in self.suites:
      return
    self.suites[suite_name] = []
    self.failure_counts[suite_name] = 0
    self.error_counts[suite_name] = 0

  def set_end_time(self, timestamp_in_secs):
    """Sets the start timestamp of this test suite.

    Args:
      timestamp_in_secs: timestamp in seconds since epoch
    """
    self.overall_end_time = timestamp_in_secs

  def set_start_time(self, timestamp_in_secs):
    """Sets the end timestamp of this test suite.

    Args:
      timestamp_in_secs: timestamp in seconds since epoch
    """
    self.overall_start_time = timestamp_in_secs


class _TextAndXMLTestResult(_pretty_print_reporter.TextTestResult):
  """Private TestResult class that produces both formatted text results and XML.

  Used by TextAndXMLTestRunner.
  """

  _TEST_SUITE_RESULT_CLASS = _TestSuiteResult
  _TEST_CASE_RESULT_CLASS = _TestCaseResult

  def __init__(self, xml_stream, stream, descriptions, verbosity,
               time_getter=_time_copy, testsuites_properties=None):
    super(_TextAndXMLTestResult, self).__init__(stream, descriptions, verbosity)
    self.xml_stream = xml_stream
    self.pending_test_case_results = {}
    self.suite = self._TEST_SUITE_RESULT_CLASS()
    if testsuites_properties:
      self.suite._testsuites_properties = testsuites_properties
    self.time_getter = time_getter

    # This lock guards any mutations on pending_test_case_results.
    self._pending_test_case_results_lock = threading.RLock()

  def startTest(self, test):
    self.start_time = self.time_getter()
    super(_TextAndXMLTestResult, self).startTest(test)

  def stopTest(self, test):
    # Grabbing the write lock to avoid conflicting with stopTestRun.
    with self._pending_test_case_results_lock:
      super(_TextAndXMLTestResult, self).stopTest(test)
      result = self.get_pending_test_case_result(test)
      if not result:
        test_name = test.id() or str(test)
        sys.stderr.write('No pending test case: %s\n' % test_name)
        return
      if getattr(self, 'start_time', None) is None:
        # startTest may not be called for skipped tests since Python 3.12.1.
        self.start_time = self.time_getter()
      test_id = id(test)
      run_time = self.time_getter() - self.start_time
      result.set_run_time(run_time)
      result.set_start_time(self.start_time)
      self.suite.add_test_case_result(result)
      del self.pending_test_case_results[test_id]

  def startTestRun(self):
    self.suite.set_start_time(self.time_getter())
    super(_TextAndXMLTestResult, self).startTestRun()

  def stopTestRun(self):
    self.suite.set_end_time(self.time_getter())
    # All pending_test_case_results will be added to the suite and removed from
    # the pending_test_case_results dictionary. Grabbing the write lock to avoid
    # results from being added during this process to avoid duplicating adds or
    # accidentally erasing newly appended pending results.
    with self._pending_test_case_results_lock:
      # Errors in the test fixture (setUpModule, tearDownModule,
      # setUpClass, tearDownClass) can leave a pending result which
      # never gets added to the suite.  The runner calls stopTestRun
      # which gives us an opportunity to add these errors for
      # reporting here.
      for test_id in self.pending_test_case_results:
        result = self.pending_test_case_results[test_id]
        if getattr(self, 'start_time', None) is not None:
          run_time = self.suite.overall_end_time - self.start_time
          result.set_run_time(run_time)
          result.set_start_time(self.start_time)
        self.suite.add_test_case_result(result)
      self.pending_test_case_results.clear()

  def _exc_info_to_string(self, err, test=None):
    """Converts a sys.exc_info()-style tuple of values into a string.

    This method must be overridden because the method signature in
    unittest.TestResult changed between Python 2.2 and 2.4.

    Args:
      err: A sys.exc_info() tuple of values for an error.
      test: The test method.

    Returns:
      A formatted exception string.
    """
    if test:
      return super(_TextAndXMLTestResult, self)._exc_info_to_string(err, test)
    return ''.join(traceback.format_exception(*err))

  def add_pending_test_case_result(self, test, error_summary=None,
                                   skip_reason=None):
    """Adds result information to a test case result which may still be running.

    If a result entry for the test already exists, add_pending_test_case_result
    will add error summary tuples and/or overwrite skip_reason for the result.
    If it does not yet exist, a result entry will be created.
    Note that a test result is considered to have been run and passed
    only if there are no errors or skip_reason.

    Args:
      test: A test method as defined by unittest
      error_summary: A 4-tuple with the following entries:
          1) a string identifier of either "failure" or "error"
          2) an exception_type
          3) an exception_message
          4) a string version of a sys.exc_info()-style tuple of values
             ('error', err[0], err[1], self._exc_info_to_string(err))
             If the length of errors is 0, then the test is either passed or
             skipped.
      skip_reason: a string explaining why the test was skipped
    """
    with self._pending_test_case_results_lock:
      test_id = id(test)
      if test_id not in self.pending_test_case_results:
        self.pending_test_case_results[test_id] = self._TEST_CASE_RESULT_CLASS(
            test)
      if error_summary:
        self.pending_test_case_results[test_id].errors.append(error_summary)
      if skip_reason:
        self.pending_test_case_results[test_id].skip_reason = skip_reason

  def delete_pending_test_case_result(self, test):
    with self._pending_test_case_results_lock:
      test_id = id(test)
      del self.pending_test_case_results[test_id]

  def get_pending_test_case_result(self, test):
    test_id = id(test)
    return self.pending_test_case_results.get(test_id, None)

  def addSuccess(self, test):
    super(_TextAndXMLTestResult, self).addSuccess(test)
    self.add_pending_test_case_result(test)

  def addError(self, test, err):
    super(_TextAndXMLTestResult, self).addError(test, err)
    error_summary = ('error', err[0], err[1],
                     self._exc_info_to_string(err, test=test))
    self.add_pending_test_case_result(test, error_summary=error_summary)

  def addFailure(self, test, err):
    super(_TextAndXMLTestResult, self).addFailure(test, err)
    error_summary = ('failure', err[0], err[1],
                     self._exc_info_to_string(err, test=test))
    self.add_pending_test_case_result(test, error_summary=error_summary)

  def addSkip(self, test, reason):
    super(_TextAndXMLTestResult, self).addSkip(test, reason)
    self.add_pending_test_case_result(test, skip_reason=reason)

  def addExpectedFailure(self, test, err):
    super(_TextAndXMLTestResult, self).addExpectedFailure(test, err)
    if callable(getattr(test, 'recordProperty', None)):
      test.recordProperty('EXPECTED_FAILURE',
                          self._exc_info_to_string(err, test=test))
    self.add_pending_test_case_result(test)

  def addUnexpectedSuccess(self, test):
    super(_TextAndXMLTestResult, self).addUnexpectedSuccess(test)
    test_name = test.id() or str(test)
    error_summary = ('error', '', '',
                     'Test case %s should have failed, but passed.'
                     % (test_name))
    self.add_pending_test_case_result(test, error_summary=error_summary)

  def addSubTest(self, test, subtest, err):  # pylint: disable=invalid-name
    super(_TextAndXMLTestResult, self).addSubTest(test, subtest, err)
    if err is not None:
      if issubclass(err[0], test.failureException):
        error_summary = ('failure', err[0], err[1],
                         self._exc_info_to_string(err, test=test))
      else:
        error_summary = ('error', err[0], err[1],
                         self._exc_info_to_string(err, test=test))
    else:
      error_summary = None
    self.add_pending_test_case_result(subtest, error_summary=error_summary)

  def printErrors(self):
    super(_TextAndXMLTestResult, self).printErrors()
    self.xml_stream.write('<?xml version="1.0"?>\n')
    self.suite.print_xml_summary(self.xml_stream)


class TextAndXMLTestRunner(unittest.TextTestRunner):
  """A test runner that produces both formatted text results and XML.

  It prints out the names of tests as they are run, errors as they
  occur, and a summary of the results at the end of the test run.
  """

  _TEST_RESULT_CLASS = _TextAndXMLTestResult

  _xml_stream = None
  _testsuites_properties = {}

  def __init__(self, xml_stream=None, *args, **kwargs):
    """Initialize a TextAndXMLTestRunner.

    Args:
      xml_stream: file-like or None; XML-formatted test results are output
          via this object's write() method.  If None (the default), the
          new instance behaves as described in the set_default_xml_stream method
          documentation below.
      *args: passed unmodified to unittest.TextTestRunner.__init__.
      **kwargs: passed unmodified to unittest.TextTestRunner.__init__.
    """
    super(TextAndXMLTestRunner, self).__init__(*args, **kwargs)
    if xml_stream is not None:
      self._xml_stream = xml_stream
    # else, do not set self._xml_stream to None -- this allows implicit fallback
    # to the class attribute's value.

  @classmethod
  def set_default_xml_stream(cls, xml_stream):
    """Sets the default XML stream for the class.

    Args:
      xml_stream: file-like or None; used for instances when xml_stream is None
          or not passed to their constructors.  If None is passed, instances
          created with xml_stream=None will act as ordinary TextTestRunner
          instances; this is the default state before any calls to this method
          have been made.
    """
    cls._xml_stream = xml_stream

  def _makeResult(self):
    if self._xml_stream is None:
      return super(TextAndXMLTestRunner, self)._makeResult()
    else:
      return self._TEST_RESULT_CLASS(
          self._xml_stream, self.stream, self.descriptions, self.verbosity,
          testsuites_properties=self._testsuites_properties)

  @classmethod
  def set_testsuites_property(cls, key, value):
    cls._testsuites_properties[key] = value



# Copyright 2017 The Abseil Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Any, Callable, Dict, NoReturn, Optional, Tuple, TypeVar, Union

from absl import flags

# Logging levels.
FATAL: int
ERROR: int
WARNING: int
WARN: int  # Deprecated name.
INFO: int
DEBUG: int

ABSL_LOGGING_PREFIX_REGEX: str

LOGTOSTDERR: flags.FlagHolder[bool]
ALSOLOGTOSTDERR: flags.FlagHolder[bool]
LOG_DIR: flags.FlagHolder[str]
VERBOSITY: flags.FlagHolder[int]
LOGGER_LEVELS: flags.FlagHolder[Dict[str, str]]
STDERRTHRESHOLD: flags.FlagHolder[str]
SHOWPREFIXFORINFO: flags.FlagHolder[bool]


def get_verbosity() -> int:
  ...


def set_verbosity(v: Union[int, str]) -> None:
  ...


def set_stderrthreshold(s: Union[int, str]) -> None:
  ...


# TODO(b/277607978): Provide actual args+kwargs shadowing stdlib's logging functions.
def fatal(msg: Any, *args: Any, **kwargs: Any) -> NoReturn:
  ...


def error(msg: Any, *args: Any, **kwargs: Any) -> None:
  ...


def warning(msg: Any, *args: Any, **kwargs: Any) -> None:
  ...


def warn(msg: Any, *args: Any, **kwargs: Any) -> None:
  ...


def info(msg: Any, *args: Any, **kwargs: Any) -> None:
  ...


def debug(msg: Any, *args: Any, **kwargs: Any) -> None:
  ...


def exception(msg: Any, *args: Any, **kwargs: Any) -> None:
  ...


def log_every_n(level: int, msg: Any, n: int, *args: Any) -> None:
  ...


def log_every_n_seconds(
    level: int, msg: Any, n_seconds: float, *args: Any
) -> None:
  ...


def log_first_n(level: int, msg: Any, n: int, *args: Any) -> None:
  ...


def log_if(level: int, msg: Any, condition: Any, *args: Any) -> None:
  ...


def log(level: int, msg: Any, *args: Any, **kwargs: Any) -> None:
  ...


def vlog(level: int, msg: Any, *args: Any, **kwargs: Any) -> None:
  ...


def vlog_is_on(level: int) -> bool:
  ...


def flush() -> None:
  ...


def level_debug() -> bool:
  ...


def level_info() -> bool:
  ...


def level_warning() -> bool:
  ...


level_warn = level_warning  # Deprecated function.


def level_error() -> bool:
  ...


def get_log_file_name(level: int = ...) -> str:
  ...


def find_log_dir_and_names(
    program_name: Optional[str] = ..., log_dir: Optional[str] = ...
) -> Tuple[str, str, str]:
  ...


def find_log_dir(log_dir: Optional[str] = ...) -> str:
  ...


def get_absl_log_prefix(record: logging.LogRecord) -> str:
  ...


_SkipLogT = TypeVar('_SkipLogT', str, Callable[..., Any])

def skip_log_prefix(func: _SkipLogT) -> _SkipLogT:
  ...


_StreamT = TypeVar("_StreamT")


class PythonHandler(logging.StreamHandler[_StreamT]):

  def __init__(
      self,
      stream: Optional[_StreamT] = ...,
      formatter: Optional[logging.Formatter] = ...,
  ) -> None:
    ...

  def start_logging_to_file(
      self, program_name: Optional[str] = ..., log_dir: Optional[str] = ...
  ) -> None:
    ...

  def use_absl_log_file(
      self, program_name: Optional[str] = ..., log_dir: Optional[str] = ...
  ) -> None:
    ...

  def flush(self) -> None:
    ...

  def emit(self, record: logging.LogRecord) -> None:
    ...

  def close(self) -> None:
    ...


class ABSLHandler(logging.Handler):

  def __init__(self, python_logging_formatter: PythonFormatter) -> None:
    ...

  def format(self, record: logging.LogRecord) -> str:
    ...

  def setFormatter(self, fmt) -> None:
    ...

  def emit(self, record: logging.LogRecord) -> None:
    ...

  def flush(self) -> None:
    ...

  def close(self) -> None:
    ...

  def handle(self, record: logging.LogRecord) -> bool:
    ...

  @property
  def python_handler(self) -> PythonHandler:
    ...

  def activate_python_handler(self) -> None:
    ...

  def use_absl_log_file(
      self, program_name: Optional[str] = ..., log_dir: Optional[str] = ...
  ) -> None:
    ...

  def start_logging_to_file(self, program_name=None, log_dir=None) -> None:
    ...


class PythonFormatter(logging.Formatter):

  def format(self, record: logging.LogRecord) -> str:
    ...


class ABSLLogger(logging.Logger):

  def findCaller(
      self, stack_info: bool = ..., stacklevel: int = ...
  ) -> Tuple[str, int, str, Optional[str]]:
    ...

  def critical(self, msg: Any, *args: Any, **kwargs: Any) -> None:
    ...

  def fatal(self, msg: Any, *args: Any, **kwargs: Any) -> NoReturn:
    ...

  def error(self, msg: Any, *args: Any, **kwargs: Any) -> None:
    ...

  def warn(self, msg: Any, *args: Any, **kwargs: Any) -> None:
    ...

  def warning(self, msg: Any, *args: Any, **kwargs: Any) -> None:
    ...

  def info(self, msg: Any, *args: Any, **kwargs: Any) -> None:
    ...

  def debug(self, msg: Any, *args: Any, **kwargs: Any) -> None:
    ...

  def log(self, level: int, msg: Any, *args: Any, **kwargs: Any) -> None:
    ...

  def handle(self, record: logging.LogRecord) -> None:
    ...

  @classmethod
  def register_frame_to_skip(
      cls, file_name: str, function_name: str, line_number: Optional[int] = ...
  ) -> None:
    ...


# NOTE: Returns None before _initialize called but shouldn't occur after import.
def get_absl_logger() -> ABSLLogger:
  ...


# NOTE: Returns None before _initialize called but shouldn't occur after import.
def get_absl_handler() -> ABSLHandler:
  ...


def use_python_logging(quiet: bool = ...) -> None:
  ...


def use_absl_handler() -> None:
  ...



# Copyright 2017 The Abseil Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for flags.FlagValues class."""

import collections
import copy
import pickle
import types
from unittest import mock

from absl import logging
from absl.flags import _defines
from absl.flags import _exceptions
from absl.flags import _flagvalues
from absl.flags import _helpers
from absl.flags import _validators
from absl.flags.tests import module_foo
from absl.testing import absltest
from absl.testing import parameterized


class FlagValuesTest(absltest.TestCase):

  def test_bool_flags(self):
    for arg, expected in (('--nothing', True),
                          ('--nothing=true', True),
                          ('--nothing=false', False),
                          ('--nonothing', False)):
      fv = _flagvalues.FlagValues()
      _defines.DEFINE_boolean('nothing', None, '', flag_values=fv)
      fv(('./program', arg))
      self.assertIs(expected, fv.nothing)

    for arg in ('--nonothing=true', '--nonothing=false'):
      fv = _flagvalues.FlagValues()
      _defines.DEFINE_boolean('nothing', None, '', flag_values=fv)
      with self.assertRaises(ValueError):
        fv(('./program', arg))

  def test_boolean_flag_parser_gets_string_argument(self):
    for arg, expected in (('--nothing', 'true'),
                          ('--nothing=true', 'true'),
                          ('--nothing=false', 'false'),
                          ('--nonothing', 'false')):
      fv = _flagvalues.FlagValues()
      _defines.DEFINE_boolean('nothing', None, '', flag_values=fv)
      with mock.patch.object(fv['nothing'].parser, 'parse') as mock_parse:
        fv(('./program', arg))
        mock_parse.assert_called_once_with(expected)

  def test_unregistered_flags_are_cleaned_up(self):
    fv = _flagvalues.FlagValues()
    module, module_name = _helpers.get_calling_module_object_and_name()

    # Define first flag.
    _defines.DEFINE_integer('cores', 4, '', flag_values=fv, short_name='c')
    old_cores_flag = fv['cores']
    fv.register_key_flag_for_module(module_name, old_cores_flag)
    self.assertEqual(fv.flags_by_module_dict(),
                     {module_name: [old_cores_flag]})
    self.assertEqual(fv.flags_by_module_id_dict(),
                     {id(module): [old_cores_flag]})
    self.assertEqual(fv.key_flags_by_module_dict(),
                     {module_name: [old_cores_flag]})

    # Redefine the same flag.
    _defines.DEFINE_integer(
        'cores', 4, '', flag_values=fv, short_name='c', allow_override=True)
    new_cores_flag = fv['cores']
    self.assertNotEqual(old_cores_flag, new_cores_flag)
    self.assertEqual(fv.flags_by_module_dict(),
                     {module_name: [new_cores_flag]})
    self.assertEqual(fv.flags_by_module_id_dict(),
                     {id(module): [new_cores_flag]})
    # old_cores_flag is removed from key flags, and the new_cores_flag is
    # not automatically added because it must be registered explicitly.
    self.assertEqual(fv.key_flags_by_module_dict(), {module_name: []})

    # Define a new flag but with the same short_name.
    _defines.DEFINE_integer(
        'changelist',
        0,
        '',
        flag_values=fv,
        short_name='c',
        allow_override=True)
    old_changelist_flag = fv['changelist']
    fv.register_key_flag_for_module(module_name, old_changelist_flag)
    # The short named flag -c is overridden to be the old_changelist_flag.
    self.assertEqual(fv['c'], old_changelist_flag)
    self.assertNotEqual(fv['c'], new_cores_flag)
    self.assertEqual(fv.flags_by_module_dict(),
                     {module_name: [new_cores_flag, old_changelist_flag]})
    self.assertEqual(fv.flags_by_module_id_dict(),
                     {id(module): [new_cores_flag, old_changelist_flag]})
    self.assertEqual(fv.key_flags_by_module_dict(),
                     {module_name: [old_changelist_flag]})

    # Define a flag only with the same long name.
    _defines.DEFINE_integer(
        'changelist',
        0,
        '',
        flag_values=fv,
        short_name='l',
        allow_override=True)
    new_changelist_flag = fv['changelist']
    self.assertNotEqual(old_changelist_flag, new_changelist_flag)
    self.assertEqual(fv.flags_by_module_dict(),
                     {module_name: [new_cores_flag,
                                    old_changelist_flag,
                                    new_changelist_flag]})
    self.assertEqual(fv.flags_by_module_id_dict(),
                     {id(module): [new_cores_flag,
                                   old_changelist_flag,
                                   new_changelist_flag]})
    self.assertEqual(fv.key_flags_by_module_dict(),
                     {module_name: [old_changelist_flag]})

    # Delete the new changelist's long name, it should still be registered
    # because of its short name.
    del fv.changelist
    self.assertNotIn('changelist', fv)
    self.assertEqual(fv.flags_by_module_dict(),
                     {module_name: [new_cores_flag,
                                    old_changelist_flag,
                                    new_changelist_flag]})
    self.assertEqual(fv.flags_by_module_id_dict(),
                     {id(module): [new_cores_flag,
                                   old_changelist_flag,
                                   new_changelist_flag]})
    self.assertEqual(fv.key_flags_by_module_dict(),
                     {module_name: [old_changelist_flag]})

    # Delete the new changelist's short name, it should be removed.
    del fv.l
    self.assertNotIn('l', fv)
    self.assertEqual(fv.flags_by_module_dict(),
                     {module_name: [new_cores_flag,
                                    old_changelist_flag]})
    self.assertEqual(fv.flags_by_module_id_dict(),
                     {id(module): [new_cores_flag,
                                   old_changelist_flag]})
    self.assertEqual(fv.key_flags_by_module_dict(),
                     {module_name: [old_changelist_flag]})

  def _test_find_module_or_id_defining_flag(self, test_id):
    """Tests for find_module_defining_flag and find_module_id_defining_flag.

    Args:
      test_id: True to test find_module_id_defining_flag, False to test
          find_module_defining_flag.
    """
    fv = _flagvalues.FlagValues()
    current_module, current_module_name = (
        _helpers.get_calling_module_object_and_name())
    alt_module_name = _flagvalues.__name__

    if test_id:
      current_module_or_id = id(current_module)
      alt_module_or_id = id(_flagvalues)
      testing_fn = fv.find_module_id_defining_flag
    else:
      current_module_or_id = current_module_name
      alt_module_or_id = alt_module_name
      testing_fn = fv.find_module_defining_flag

    # Define first flag.
    _defines.DEFINE_integer('cores', 4, '', flag_values=fv, short_name='c')
    module_or_id_cores = testing_fn('cores')
    self.assertEqual(module_or_id_cores, current_module_or_id)
    module_or_id_c = testing_fn('c')
    self.assertEqual(module_or_id_c, current_module_or_id)

    # Redefine the same flag in another module.
    _defines.DEFINE_integer(
        'cores',
        4,
        '',
        flag_values=fv,
        module_name=alt_module_name,
        short_name='c',
        allow_override=True)
    module_or_id_cores = testing_fn('cores')
    self.assertEqual(module_or_id_cores, alt_module_or_id)
    module_or_id_c = testing_fn('c')
    self.assertEqual(module_or_id_c, alt_module_or_id)

    # Define a new flag but with the same short_name.
    _defines.DEFINE_integer(
        'changelist',
        0,
        '',
        flag_values=fv,
        short_name='c',
        allow_override=True)
    module_or_id_cores = testing_fn('cores')
    self.assertEqual(module_or_id_cores, alt_module_or_id)
    module_or_id_changelist = testing_fn('changelist')
    self.assertEqual(module_or_id_changelist, current_module_or_id)
    module_or_id_c = testing_fn('c')
    self.assertEqual(module_or_id_c, current_module_or_id)

    # Define a flag in another module only with the same long name.
    _defines.DEFINE_integer(
        'changelist',
        0,
        '',
        flag_values=fv,
        module_name=alt_module_name,
        short_name='l',
        allow_override=True)
    module_or_id_cores = testing_fn('cores')
    self.assertEqual(module_or_id_cores, alt_module_or_id)
    module_or_id_changelist = testing_fn('changelist')
    self.assertEqual(module_or_id_changelist, alt_module_or_id)
    module_or_id_c = testing_fn('c')
    self.assertEqual(module_or_id_c, current_module_or_id)
    module_or_id_l = testing_fn('l')
    self.assertEqual(module_or_id_l, alt_module_or_id)

    # Delete the changelist flag, its short name should still be registered.
    del fv.changelist
    module_or_id_changelist = testing_fn('changelist')
    self.assertIsNone(module_or_id_changelist)
    module_or_id_c = testing_fn('c')
    self.assertEqual(module_or_id_c, current_module_or_id)
    module_or_id_l = testing_fn('l')
    self.assertEqual(module_or_id_l, alt_module_or_id)

  def test_find_module_defining_flag(self):
    self._test_find_module_or_id_defining_flag(test_id=False)

  def test_find_module_id_defining_flag(self):
    self._test_find_module_or_id_defining_flag(test_id=True)

  def test_set_default(self):
    fv = _flagvalues.FlagValues()
    fv.mark_as_parsed()
    with self.assertRaises(_exceptions.UnrecognizedFlagError):
      fv.set_default('changelist', 1)
    _defines.DEFINE_integer('changelist', 0, 'help', flag_values=fv)
    self.assertEqual(0, fv.changelist)
    fv.set_default('changelist', 2)
    self.assertEqual(2, fv.changelist)

  def test_default_gnu_getopt_value(self):
    self.assertTrue(_flagvalues.FlagValues().is_gnu_getopt())

  def test_known_only_flags_in_gnustyle(self):

    def run_test(argv, defined_py_flags, expected_argv):
      fv = _flagvalues.FlagValues()
      fv.set_gnu_getopt(True)
      for f in defined_py_flags:
        if f.startswith('b'):
          _defines.DEFINE_boolean(f, False, 'help', flag_values=fv)
        else:
          _defines.DEFINE_string(f, 'default', 'help', flag_values=fv)
      output_argv = fv(argv, known_only=True)
      self.assertEqual(expected_argv, output_argv)

    run_test(
        argv='0 --f1=v1 cmd --f2 v2 --b1 --f3 v3 --nob2'.split(' '),
        defined_py_flags=[],
        expected_argv='0 --f1=v1 cmd --f2 v2 --b1 --f3 v3 --nob2'.split(' '))
    run_test(
        argv='0 --f1=v1 cmd --f2 v2 --b1 --f3 v3 --nob2'.split(' '),
        defined_py_flags=['f1'],
        expected_argv='0 cmd --f2 v2 --b1 --f3 v3 --nob2'.split(' '))
    run_test(
        argv='0 --f1=v1 cmd --f2 v2 --b1 --f3 v3 --nob2'.split(' '),
        defined_py_flags=['f2'],
        expected_argv='0 --f1=v1 cmd --b1 --f3 v3 --nob2'.split(' '))
    run_test(
        argv='0 --f1=v1 cmd --f2 v2 --b1 --f3 v3 --nob2'.split(' '),
        defined_py_flags=['b1'],
        expected_argv='0 --f1=v1 cmd --f2 v2 --f3 v3 --nob2'.split(' '))
    run_test(
        argv='0 --f1=v1 cmd --f2 v2 --b1 --f3 v3 --nob2'.split(' '),
        defined_py_flags=['f3'],
        expected_argv='0 --f1=v1 cmd --f2 v2 --b1 --nob2'.split(' '))
    run_test(
        argv='0 --f1=v1 cmd --f2 v2 --b1 --f3 v3 --nob2'.split(' '),
        defined_py_flags=['b2'],
        expected_argv='0 --f1=v1 cmd --f2 v2 --b1 --f3 v3'.split(' '))
    run_test(
        argv=('0 --f1=v1 cmd --undefok=f1 --f2 v2 --b1 '
              '--f3 v3 --nob2').split(' '),
        defined_py_flags=['b2'],
        expected_argv='0 cmd --f2 v2 --b1 --f3 v3'.split(' '))
    run_test(
        argv=('0 --f1=v1 cmd --undefok f1,f2 --f2 v2 --b1 '
              '--f3 v3 --nob2').split(' '),
        defined_py_flags=['b2'],
        # Note v2 is preserved here, since undefok requires the flag being
        # specified in the form of --flag=value.
        expected_argv='0 cmd v2 --b1 --f3 v3'.split(' '))

  def test_invalid_flag_name(self):
    with self.assertRaises(_exceptions.Error):
      _defines.DEFINE_boolean('test ', 0, '')

    with self.assertRaises(_exceptions.Error):
      _defines.DEFINE_boolean(' test', 0, '')

    with self.assertRaises(_exceptions.Error):
      _defines.DEFINE_boolean('te st', 0, '')

    with self.assertRaises(_exceptions.Error):
      _defines.DEFINE_boolean('', 0, '')

    with self.assertRaises(_exceptions.Error):
      _defines.DEFINE_boolean(1, 0, '')  # type: ignore

  def test_len(self):
    fv = _flagvalues.FlagValues()
    self.assertEmpty(fv)
    self.assertFalse(fv)

    _defines.DEFINE_boolean('boolean', False, 'help', flag_values=fv)
    self.assertLen(fv, 1)
    self.assertTrue(fv)

    _defines.DEFINE_boolean(
        'bool', False, 'help', short_name='b', flag_values=fv)
    self.assertLen(fv, 3)
    self.assertTrue(fv)

  def test_pickle(self):
    fv = _flagvalues.FlagValues()
    with self.assertRaisesRegex(TypeError, "can't pickle FlagValues"):
      pickle.dumps(fv)

  def test_copy(self):
    fv = _flagvalues.FlagValues()
    _defines.DEFINE_integer('answer', 0, 'help', flag_values=fv)
    fv(['', '--answer=1'])

    with self.assertRaisesRegex(TypeError,
                                'FlagValues does not support shallow copies'):
      copy.copy(fv)

    fv2 = copy.deepcopy(fv)
    self.assertEqual(fv2.answer, 1)

    fv2.answer = 42
    self.assertEqual(fv2.answer, 42)
    self.assertEqual(fv.answer, 1)

  def test_conflicting_flags(self):
    fv = _flagvalues.FlagValues()
    with self.assertRaises(_exceptions.FlagNameConflictsWithMethodError):
      _defines.DEFINE_boolean('is_gnu_getopt', False, 'help', flag_values=fv)
    _defines.DEFINE_boolean(
        'is_gnu_getopt',
        False,
        'help',
        flag_values=fv,
        allow_using_method_names=True)
    self.assertFalse(fv['is_gnu_getopt'].value)
    self.assertIsInstance(fv.is_gnu_getopt, types.MethodType)

  def test_get_flags_for_module(self):
    fv = _flagvalues.FlagValues()
    _defines.DEFINE_string('foo', None, 'help', flag_values=fv)
    module_foo.define_flags(fv)
    flags = fv.get_flags_for_module('__main__')

    self.assertEqual({'foo'}, {flag.name for flag in flags})

    flags = fv.get_flags_for_module(module_foo)
    self.assertEqual({'tmod_foo_bool', 'tmod_foo_int', 'tmod_foo_str'},
                     {flag.name for flag in flags})

  def test_get_help(self):
    fv = _flagvalues.FlagValues()
    self.assertMultiLineEqual('''\
--flagfile: Insert flag definitions from the given file into the command line.
  (default: '')
--undefok: comma-separated list of flag names that it is okay to specify on the
  command line even if the program does not define a flag with that name.
  IMPORTANT: flags in this list that have arguments MUST use the --flag=value
  format.
  (default: '')''', fv.get_help())

    module_foo.define_flags(fv)
    self.assertMultiLineEqual('''
absl.flags.tests.module_bar:
  --tmod_bar_t: Sample int flag.
    (default: '4')
    (an integer)
  --tmod_bar_u: Sample int flag.
    (default: '5')
    (an integer)
  --tmod_bar_v: Sample int flag.
    (default: '6')
    (an integer)
  --[no]tmod_bar_x: Boolean flag.
    (default: 'true')
  --tmod_bar_y: String flag.
    (default: 'default')
  --[no]tmod_bar_z: Another boolean flag from module bar.
    (default: 'false')

absl.flags.tests.module_foo:
  --[no]tmod_foo_bool: Boolean flag from module foo.
    (default: 'true')
  --tmod_foo_int: Sample int flag.
    (default: '3')
    (an integer)
  --tmod_foo_str: String flag.
    (default: 'default')

absl.flags:
  --flagfile: Insert flag definitions from the given file into the command line.
    (default: '')
  --undefok: comma-separated list of flag names that it is okay to specify on
    the command line even if the program does not define a flag with that name.
    IMPORTANT: flags in this list that have arguments MUST use the --flag=value
    format.
    (default: '')''', fv.get_help())

    self.assertMultiLineEqual('''
xxxxabsl.flags.tests.module_bar:
xxxx  --tmod_bar_t: Sample int flag.
xxxx    (default: '4')
xxxx    (an integer)
xxxx  --tmod_bar_u: Sample int flag.
xxxx    (default: '5')
xxxx    (an integer)
xxxx  --tmod_bar_v: Sample int flag.
xxxx    (default: '6')
xxxx    (an integer)
xxxx  --[no]tmod_bar_x: Boolean flag.
xxxx    (default: 'true')
xxxx  --tmod_bar_y: String flag.
xxxx    (default: 'default')
xxxx  --[no]tmod_bar_z: Another boolean flag from module bar.
xxxx    (default: 'false')

xxxxabsl.flags.tests.module_foo:
xxxx  --[no]tmod_foo_bool: Boolean flag from module foo.
xxxx    (default: 'true')
xxxx  --tmod_foo_int: Sample int flag.
xxxx    (default: '3')
xxxx    (an integer)
xxxx  --tmod_foo_str: String flag.
xxxx    (default: 'default')

xxxxabsl.flags:
xxxx  --flagfile: Insert flag definitions from the given file into the command
xxxx    line.
xxxx    (default: '')
xxxx  --undefok: comma-separated list of flag names that it is okay to specify
xxxx    on the command line even if the program does not define a flag with that
xxxx    name.  IMPORTANT: flags in this list that have arguments MUST use the
xxxx    --flag=value format.
xxxx    (default: '')''', fv.get_help(prefix='xxxx'))

    self.assertMultiLineEqual('''
absl.flags.tests.module_bar:
  --tmod_bar_t: Sample int flag.
    (default: '4')
    (an integer)
  --tmod_bar_u: Sample int flag.
    (default: '5')
    (an integer)
  --tmod_bar_v: Sample int flag.
    (default: '6')
    (an integer)
  --[no]tmod_bar_x: Boolean flag.
    (default: 'true')
  --tmod_bar_y: String flag.
    (default: 'default')
  --[no]tmod_bar_z: Another boolean flag from module bar.
    (default: 'false')

absl.flags.tests.module_foo:
  --[no]tmod_foo_bool: Boolean flag from module foo.
    (default: 'true')
  --tmod_foo_int: Sample int flag.
    (default: '3')
    (an integer)
  --tmod_foo_str: String flag.
    (default: 'default')''', fv.get_help(include_special_flags=False))

  def test_str(self):
    fv = _flagvalues.FlagValues()
    self.assertEqual(str(fv), fv.get_help())
    module_foo.define_flags(fv)
    self.assertEqual(str(fv), fv.get_help())

  def test_empty_argv(self):
    fv = _flagvalues.FlagValues()
    with self.assertRaises(ValueError):
      fv([])

  def test_invalid_argv(self):
    fv = _flagvalues.FlagValues()
    with self.assertRaises(TypeError):
      fv('./program')  # type: ignore
    with self.assertRaises(TypeError):
      fv(b'./program')  # type: ignore

  def test_flags_dir(self):
    flag_values = _flagvalues.FlagValues()
    flag_name1 = 'bool_flag'
    flag_name2 = 'string_flag'
    flag_name3 = 'float_flag'
    description = 'Description'
    _defines.DEFINE_boolean(
        flag_name1, None, description, flag_values=flag_values)
    _defines.DEFINE_string(
        flag_name2, None, description, flag_values=flag_values)
    self.assertEqual(sorted([flag_name1, flag_name2]), dir(flag_values))

    _defines.DEFINE_float(
        flag_name3, None, description, flag_values=flag_values)
    self.assertEqual(
        sorted([flag_name1, flag_name2, flag_name3]), dir(flag_values))

  def test_flags_into_string_deterministic(self):
    flag_values = _flagvalues.FlagValues()
    _defines.DEFINE_string(
        'fa', 'x', '', flag_values=flag_values, module_name='mb')
    _defines.DEFINE_string(
        'fb', 'x', '', flag_values=flag_values, module_name='mb')
    _defines.DEFINE_string(
        'fc', 'x', '', flag_values=flag_values, module_name='ma')
    _defines.DEFINE_string(
        'fd', 'x', '', flag_values=flag_values, module_name='ma')

    expected = ('--fc=x\n'
                '--fd=x\n'
                '--fa=x\n'
                '--fb=x\n')

    flags_by_module_items = sorted(
        flag_values.flags_by_module_dict().items(), reverse=True)
    for _, module_flags in flags_by_module_items:
      module_flags.sort(reverse=True)

    flag_values.__dict__['__flags_by_module'] = collections.OrderedDict(
        flags_by_module_items)

    actual = flag_values.flags_into_string()
    self.assertEqual(expected, actual)

  def test_validate_all_flags(self):
    fv = _flagvalues.FlagValues()
    _defines.DEFINE_string('name', None, '', flag_values=fv)
    _validators.mark_flag_as_required('name', flag_values=fv)
    with self.assertRaises(_exceptions.IllegalFlagValueError):
      fv.validate_all_flags()
    fv.name = 'test'
    fv.validate_all_flags()


class FlagValuesLoggingTest(absltest.TestCase):
  """Test to make sure logging.* functions won't recurse.

  Logging may and does happen before flags initialization. We need to make
  sure that any warnings trown by flagvalues do not result in unlimited
  recursion.
  """

  def test_logging_do_not_recurse(self):
    logging.info('test info')
    try:
      raise ValueError('test exception')
    except ValueError:
      logging.exception('test message')


class FlagSubstrMatchingTests(parameterized.TestCase):
  """Tests related to flag substring matching."""

  def _get_test_flag_values(self):
    """Get a _flagvalues.FlagValues() instance, set up for tests."""
    flag_values = _flagvalues.FlagValues()

    _defines.DEFINE_string('strf', '', '', flag_values=flag_values)
    _defines.DEFINE_boolean('boolf', 0, '', flag_values=flag_values)

    return flag_values

  # Test cases that should always make parsing raise an error.
  # Tuples of strings with the argv to use.
  FAIL_TEST_CASES = [
      ('./program', '--boo', '0'),
      ('./program', '--boo=true', '0'),
      ('./program', '--boo=0'),
      ('./program', '--noboo'),
      ('./program', '--st=blah'),
      ('./program', '--st=de'),
      ('./program', '--st=blah', '--boo'),
      ('./program', '--st=blah', 'unused'),
      ('./program', '--st=--blah'),
      ('./program', '--st', '--blah'),
  ]

  @parameterized.parameters(FAIL_TEST_CASES)
  def test_raise(self, *argv):
    """Test that raising works."""
    fv = self._get_test_flag_values()
    with self.assertRaises(_exceptions.UnrecognizedFlagError):
      fv(argv)

  @parameterized.parameters(
      FAIL_TEST_CASES + [('./program', 'unused', '--st=blah')])
  def test_gnu_getopt_raise(self, *argv):
    """Test that raising works when combined with GNU-style getopt."""
    fv = self._get_test_flag_values()
    fv.set_gnu_getopt()
    with self.assertRaises(_exceptions.UnrecognizedFlagError):
      fv(argv)


class SettingUnknownFlagTest(absltest.TestCase):

  def setUp(self):
    super(SettingUnknownFlagTest, self).setUp()
    self.setter_called = 0

  def set_undef(self, unused_name, unused_val):
    self.setter_called += 1

  def test_raise_on_undefined(self):
    new_flags = _flagvalues.FlagValues()
    with self.assertRaises(_exceptions.UnrecognizedFlagError):
      new_flags.undefined_flag = 0

  def test_not_raise(self):
    new_flags = _flagvalues.FlagValues()
    new_flags._register_unknown_flag_setter(self.set_undef)
    new_flags.undefined_flag = 0
    self.assertEqual(self.setter_called, 1)

  def test_not_raise_on_undefined_if_undefok(self):
    new_flags = _flagvalues.FlagValues()
    args = ['0', '--foo', '--bar=1', '--undefok=foo,bar']
    unparsed = new_flags(args, known_only=True)
    self.assertEqual(['0'], unparsed)

  def test_re_raise_undefined(self):
    def setter(unused_name, unused_val):
      raise NameError()
    new_flags = _flagvalues.FlagValues()
    new_flags._register_unknown_flag_setter(setter)
    with self.assertRaises(_exceptions.UnrecognizedFlagError):
      new_flags.undefined_flag = 0

  def test_re_raise_invalid(self):
    def setter(unused_name, unused_val):
      raise ValueError()
    new_flags = _flagvalues.FlagValues()
    new_flags._register_unknown_flag_setter(setter)
    with self.assertRaises(_exceptions.IllegalFlagValueError):
      new_flags.undefined_flag = 0


class SetAttributesTest(absltest.TestCase):

  def setUp(self):
    super(SetAttributesTest, self).setUp()
    self.new_flags = _flagvalues.FlagValues()
    _defines.DEFINE_boolean(
        'defined_flag', None, '', flag_values=self.new_flags)
    _defines.DEFINE_boolean(
        'another_defined_flag', None, '', flag_values=self.new_flags)
    self.setter_called = 0

  def set_undef(self, unused_name, unused_val):
    self.setter_called += 1

  def test_two_defined_flags(self):
    self.new_flags._set_attributes(
        defined_flag=False, another_defined_flag=False)
    self.assertEqual(self.setter_called, 0)

  def test_one_defined_one_undefined_flag(self):
    with self.assertRaises(_exceptions.UnrecognizedFlagError):
      self.new_flags._set_attributes(defined_flag=False, undefined_flag=0)

  def test_register_unknown_flag_setter(self):
    self.new_flags._register_unknown_flag_setter(self.set_undef)
    self.new_flags._set_attributes(defined_flag=False, undefined_flag=0)
    self.assertEqual(self.setter_called, 1)


class FlagsDashSyntaxTest(absltest.TestCase):

  def setUp(self):
    super(FlagsDashSyntaxTest, self).setUp()
    self.fv = _flagvalues.FlagValues()
    _defines.DEFINE_string(
        'long_name', 'default', 'help', flag_values=self.fv, short_name='s')

  def test_long_name_one_dash(self):
    self.fv(['./program', '-long_name=new'])
    self.assertEqual('new', self.fv.long_name)

  def test_long_name_two_dashes(self):
    self.fv(['./program', '--long_name=new'])
    self.assertEqual('new', self.fv.long_name)

  def test_long_name_three_dashes(self):
    with self.assertRaises(_exceptions.UnrecognizedFlagError):
      self.fv(['./program', '---long_name=new'])

  def test_short_name_one_dash(self):
    self.fv(['./program', '-s=new'])
    self.assertEqual('new', self.fv.s)

  def test_short_name_two_dashes(self):
    self.fv(['./program', '--s=new'])
    self.assertEqual('new', self.fv.s)

  def test_short_name_three_dashes(self):
    with self.assertRaises(_exceptions.UnrecognizedFlagError):
      self.fv(['./program', '---s=new'])


class UnparseFlagsTest(absltest.TestCase):

  def test_using_default_value_none(self):
    fv = _flagvalues.FlagValues()
    _defines.DEFINE_string('default_none', None, 'help', flag_values=fv)
    self.assertTrue(fv['default_none'].using_default_value)
    fv(['', '--default_none=notNone'])
    self.assertFalse(fv['default_none'].using_default_value)
    fv.unparse_flags()
    self.assertTrue(fv['default_none'].using_default_value)
    fv(['', '--default_none=alsoNotNone'])
    self.assertFalse(fv['default_none'].using_default_value)
    fv.unparse_flags()
    self.assertTrue(fv['default_none'].using_default_value)

  def test_using_default_value_not_none(self):
    fv = _flagvalues.FlagValues()
    _defines.DEFINE_string('default_foo', 'foo', 'help', flag_values=fv)

    fv.mark_as_parsed()
    self.assertTrue(fv['default_foo'].using_default_value)

    fv(['', '--default_foo=foo'])
    self.assertFalse(fv['default_foo'].using_default_value)

    fv(['', '--default_foo=notFoo'])
    self.assertFalse(fv['default_foo'].using_default_value)

    fv.unparse_flags()
    self.assertTrue(fv['default_foo'].using_default_value)

    fv(['', '--default_foo=alsoNotFoo'])
    self.assertFalse(fv['default_foo'].using_default_value)

  def test_allow_overwrite_false(self):
    fv = _flagvalues.FlagValues()
    _defines.DEFINE_string(
        'default_none', None, 'help', allow_overwrite=False, flag_values=fv)
    _defines.DEFINE_string(
        'default_foo', 'foo', 'help', allow_overwrite=False, flag_values=fv)

    fv.mark_as_parsed()
    self.assertEqual('foo', fv.default_foo)
    self.assertIsNone(fv.default_none)

    fv(['', '--default_foo=notFoo', '--default_none=notNone'])
    self.assertEqual('notFoo', fv.default_foo)
    self.assertEqual('notNone', fv.default_none)

    fv.unparse_flags()
    self.assertEqual('foo', fv['default_foo'].value)
    self.assertIsNone(fv['default_none'].value)

    fv(['', '--default_foo=alsoNotFoo', '--default_none=alsoNotNone'])
    self.assertEqual('alsoNotFoo', fv.default_foo)
    self.assertEqual('alsoNotNone', fv.default_none)

  def test_multi_string_default_none(self):
    fv = _flagvalues.FlagValues()
    _defines.DEFINE_multi_string('foo', None, 'help', flag_values=fv)
    fv.mark_as_parsed()
    self.assertIsNone(fv.foo)
    fv(['', '--foo=aa'])
    self.assertEqual(['aa'], fv.foo)
    fv.unparse_flags()
    self.assertIsNone(fv['foo'].value)
    fv(['', '--foo=bb', '--foo=cc'])
    self.assertEqual(['bb', 'cc'], fv.foo)
    fv.unparse_flags()
    self.assertIsNone(fv['foo'].value)

  def test_multi_string_default_string(self):
    fv = _flagvalues.FlagValues()
    _defines.DEFINE_multi_string('foo', 'xyz', 'help', flag_values=fv)
    expected_default = ['xyz']
    fv.mark_as_parsed()
    self.assertEqual(expected_default, fv.foo)
    fv(['', '--foo=aa'])
    self.assertEqual(['aa'], fv.foo)
    fv.unparse_flags()
    self.assertEqual(expected_default, fv['foo'].value)
    fv(['', '--foo=bb', '--foo=cc'])
    self.assertEqual(['bb', 'cc'], fv['foo'].value)
    fv.unparse_flags()
    self.assertEqual(expected_default, fv['foo'].value)

  def test_multi_string_default_list(self):
    fv = _flagvalues.FlagValues()
    _defines.DEFINE_multi_string(
        'foo', ['xx', 'yy', 'zz'], 'help', flag_values=fv)
    expected_default = ['xx', 'yy', 'zz']
    fv.mark_as_parsed()
    self.assertEqual(expected_default, fv.foo)
    fv(['', '--foo=aa'])
    self.assertEqual(['aa'], fv.foo)
    fv.unparse_flags()
    self.assertEqual(expected_default, fv['foo'].value)
    fv(['', '--foo=bb', '--foo=cc'])
    self.assertEqual(['bb', 'cc'], fv.foo)
    fv.unparse_flags()
    self.assertEqual(expected_default, fv['foo'].value)


class UnparsedFlagAccessTest(absltest.TestCase):

  def test_unparsed_flag_access(self):
    fv = _flagvalues.FlagValues()
    _defines.DEFINE_string('name', 'default', 'help', flag_values=fv)
    with self.assertRaises(_exceptions.UnparsedFlagAccessError):
      _ = fv.name

  def test_hasattr_raises_in_py3(self):
    fv = _flagvalues.FlagValues()
    _defines.DEFINE_string('name', 'default', 'help', flag_values=fv)
    with self.assertRaises(_exceptions.UnparsedFlagAccessError):
      _ = hasattr(fv, 'name')

  def test_unparsed_flags_access_raises_after_unparse_flags(self):
    fv = _flagvalues.FlagValues()
    _defines.DEFINE_string('a_str', 'default_value', 'help', flag_values=fv)
    fv.mark_as_parsed()
    self.assertEqual(fv.a_str, 'default_value')
    fv.unparse_flags()
    with self.assertRaises(_exceptions.UnparsedFlagAccessError):
      _ = fv.a_str


class FlagHolderTest(absltest.TestCase):

  def setUp(self):
    super(FlagHolderTest, self).setUp()
    self.fv = _flagvalues.FlagValues()
    self.name_flag = _defines.DEFINE_string(
        'name', 'default', 'help', flag_values=self.fv)

  def parse_flags(self, *argv):
    self.fv.unparse_flags()
    self.fv(['binary_name'] + list(argv))

  def test_name(self):
    self.assertEqual('name', self.name_flag.name)

  def test_value_before_flag_parsing(self):
    with self.assertRaises(_exceptions.UnparsedFlagAccessError):
      _ = self.name_flag.value

  def test_value_returns_default_value_if_not_explicitly_set(self):
    self.parse_flags()
    self.assertEqual('default', self.name_flag.value)

  def test_value_returns_explicitly_set_value(self):
    self.parse_flags('--name=new_value')
    self.assertEqual('new_value', self.name_flag.value)

  def test_present_returns_false_before_flag_parsing(self):
    self.assertFalse(self.name_flag.present)

  def test_present_returns_false_if_not_explicitly_set(self):
    self.parse_flags()
    self.assertFalse(self.name_flag.present)

  def test_present_returns_true_if_explicitly_set(self):
    self.parse_flags('--name=new_value')
    self.assertTrue(self.name_flag.present)

  def test_serializes_flag(self):
    self.parse_flags('--name=new_value')
    self.assertEqual('--name=new_value', self.name_flag.serialize())

  def test_allow_override(self):
    first = _defines.DEFINE_integer(
        'int_flag', 1, 'help', flag_values=self.fv, allow_override=1)
    second = _defines.DEFINE_integer(
        'int_flag', 2, 'help', flag_values=self.fv, allow_override=1)
    self.parse_flags('--int_flag=3')
    self.assertEqual(3, first.value)
    self.assertEqual(3, second.value)
    self.assertTrue(first.present)
    self.assertTrue(second.present)

  def test_eq(self):
    with self.assertRaises(TypeError):
      self.name_flag == 'value'  # pylint: disable=pointless-statement

  def test_eq_reflection(self):
    with self.assertRaises(TypeError):
      'value' == self.name_flag  # pylint: disable=pointless-statement

  def test_bool(self):
    with self.assertRaises(TypeError):
      bool(self.name_flag)


if __name__ == '__main__':
  absltest.main()



# Copyright 2017 The Abseil Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unittests for helpers module."""

import sys

from absl.flags import _helpers
from absl.flags.tests import module_bar
from absl.flags.tests import module_foo
from absl.testing import absltest


class FlagSuggestionTest(absltest.TestCase):

  def setUp(self):
    self.longopts = [
        'fsplit-ivs-in-unroller=',
        'fsplit-wide-types=',
        'fstack-protector=',
        'fstack-protector-all=',
        'fstrict-aliasing=',
        'fstrict-overflow=',
        'fthread-jumps=',
        'ftracer',
        'ftree-bit-ccp',
        'ftree-builtin-call-dce',
        'ftree-ccp',
        'ftree-ch']

  def test_damerau_levenshtein_id(self):
    self.assertEqual(0, _helpers._damerau_levenshtein('asdf', 'asdf'))

  def test_damerau_levenshtein_empty(self):
    self.assertEqual(5, _helpers._damerau_levenshtein('', 'kites'))
    self.assertEqual(6, _helpers._damerau_levenshtein('kitten', ''))

  def test_damerau_levenshtein_commutative(self):
    self.assertEqual(2, _helpers._damerau_levenshtein('kitten', 'kites'))
    self.assertEqual(2, _helpers._damerau_levenshtein('kites', 'kitten'))

  def test_damerau_levenshtein_transposition(self):
    self.assertEqual(1, _helpers._damerau_levenshtein('kitten', 'ktiten'))

  def test_mispelled_suggestions(self):
    suggestions = _helpers.get_flag_suggestions('fstack_protector_all',
                                                self.longopts)
    self.assertEqual(['fstack-protector-all'], suggestions)

  def test_ambiguous_prefix_suggestion(self):
    suggestions = _helpers.get_flag_suggestions('fstack', self.longopts)
    self.assertEqual(['fstack-protector', 'fstack-protector-all'], suggestions)

  def test_misspelled_ambiguous_prefix_suggestion(self):
    suggestions = _helpers.get_flag_suggestions('stack', self.longopts)
    self.assertEqual(['fstack-protector', 'fstack-protector-all'], suggestions)

  def test_crazy_suggestion(self):
    suggestions = _helpers.get_flag_suggestions('asdfasdgasdfa', self.longopts)
    self.assertEqual([], suggestions)

  def test_suggestions_are_sorted(self):
    sorted_flags = sorted(['aab', 'aac', 'aad'])
    misspelt_flag = 'aaa'
    suggestions = _helpers.get_flag_suggestions(
        misspelt_flag, list(reversed(sorted_flags))
    )
    self.assertEqual(sorted_flags, suggestions)


class GetCallingModuleTest(absltest.TestCase):
  """Test whether we correctly determine the module which defines the flag."""

  def test_get_calling_module(self):
    self.assertEqual(_helpers.get_calling_module(), sys.argv[0])
    self.assertEqual(module_foo.get_module_name(),
                     'absl.flags.tests.module_foo')
    self.assertEqual(module_bar.get_module_name(),
                     'absl.flags.tests.module_bar')

    # We execute the following exec statements for their side-effect
    # (i.e., not raising an error).  They emphasize the case that not
    # all code resides in one of the imported modules: Python is a
    # really dynamic language, where we can dynamically construct some
    # code and execute it.
    code = ('from absl.flags import _helpers\n'
            'module_name = _helpers.get_calling_module()')
    exec(code)  # pylint: disable=exec-used

    # Next two exec statements executes code with a global environment
    # that is different from the global environment of any imported
    # module.
    exec(code, {})  # pylint: disable=exec-used
    # vars(self) returns a dictionary corresponding to the symbol
    # table of the self object.  dict(...) makes a distinct copy of
    # this dictionary, such that any new symbol definition by the
    # exec-ed code (e.g., import flags, module_name = ...) does not
    # affect the symbol table of self.
    exec(code, dict(vars(self)))  # pylint: disable=exec-used

    # Next test is actually more involved: it checks not only that
    # get_calling_module does not crash inside exec code, it also checks
    # that it returns the expected value: the code executed via exec
    # code is treated as being executed by the current module.  We
    # check it twice: first time by executing exec from the main
    # module, second time by executing it from module_bar.
    global_dict = {}
    exec(code, global_dict)  # pylint: disable=exec-used
    self.assertEqual(global_dict['module_name'],
                     sys.argv[0])

    global_dict = {}
    module_bar.execute_code(code, global_dict)
    self.assertEqual(global_dict['module_name'],
                     'absl.flags.tests.module_bar')

  def test_get_calling_module_with_iteritems_error(self):
    # This test checks that get_calling_module is using
    # sys.modules.items(), instead of .iteritems().
    orig_sys_modules = sys.modules

    # Mock sys.modules: simulates error produced by importing a module
    # in parallel with our iteration over sys.modules.iteritems().
    class SysModulesMock(dict):

      def __init__(self, original_content):
        dict.__init__(self, original_content)

      def iteritems(self):
        # Any dictionary method is fine, but not .iteritems().
        raise RuntimeError('dictionary changed size during iteration')

    sys.modules = SysModulesMock(orig_sys_modules)
    try:
      # _get_calling_module should still work as expected:
      self.assertEqual(_helpers.get_calling_module(), sys.argv[0])
      self.assertEqual(module_foo.get_module_name(),
                       'absl.flags.tests.module_foo')
    finally:
      sys.modules = orig_sys_modules


if __name__ == '__main__':
  absltest.main()



# Copyright 2017 The Abseil Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl import flags
from absl.flags import _helpers
from absl.testing import absltest

FLAGS = flags.FLAGS


class FlagsUnitTest(absltest.TestCase):
  """Flags formatting Unit Test."""

  def test_get_help_width(self):
    """Verify that get_help_width() reflects _help_width."""
    default_help_width = _helpers._DEFAULT_HELP_WIDTH  # Save.
    self.assertEqual(80, _helpers._DEFAULT_HELP_WIDTH)
    self.assertEqual(_helpers._DEFAULT_HELP_WIDTH, flags.get_help_width())
    _helpers._DEFAULT_HELP_WIDTH = 10
    self.assertEqual(_helpers._DEFAULT_HELP_WIDTH, flags.get_help_width())
    _helpers._DEFAULT_HELP_WIDTH = default_help_width  # restore

  def test_text_wrap(self):
    """Test that wrapping works as expected.

    Also tests that it is using global flags._help_width by default.
    """
    default_help_width = _helpers._DEFAULT_HELP_WIDTH
    _helpers._DEFAULT_HELP_WIDTH = 10

    # Generate a string with length 40, no spaces
    text = ''
    expect = []
    for n in range(4):
      line = str(n)
      line += '123456789'
      text += line
      expect.append(line)

    # Verify we still break
    wrapped = flags.text_wrap(text).split('\n')
    self.assertEqual(4, len(wrapped))
    self.assertEqual(expect, wrapped)

    wrapped = flags.text_wrap(text, 80).split('\n')
    self.assertEqual(1, len(wrapped))
    self.assertEqual([text], wrapped)

    # Normal case, breaking at word boundaries and rewriting new lines
    input_value = 'a b c d e f g h'
    expect = {1: ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h'],
              2: ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h'],
              3: ['a b', 'c d', 'e f', 'g h'],
              4: ['a b', 'c d', 'e f', 'g h'],
              5: ['a b c', 'd e f', 'g h'],
              6: ['a b c', 'd e f', 'g h'],
              7: ['a b c d', 'e f g h'],
              8: ['a b c d', 'e f g h'],
              9: ['a b c d e', 'f g h'],
              10: ['a b c d e', 'f g h'],
              11: ['a b c d e f', 'g h'],
              12: ['a b c d e f', 'g h'],
              13: ['a b c d e f g', 'h'],
              14: ['a b c d e f g', 'h'],
              15: ['a b c d e f g h']}
    for width, exp in expect.items():
      self.assertEqual(exp, flags.text_wrap(input_value, width).split('\n'))

    # We turn lines with only whitespace into empty lines
    # We strip from the right up to the first new line
    self.assertEqual('', flags.text_wrap('  '))
    self.assertEqual('\n', flags.text_wrap('  \n  '))
    self.assertEqual('\n', flags.text_wrap('\n\n'))
    self.assertEqual('\n\n', flags.text_wrap('\n\n\n'))
    self.assertEqual('\n', flags.text_wrap('\n '))
    self.assertEqual('a\n\nb', flags.text_wrap('a\n  \nb'))
    self.assertEqual('a\n\n\nb', flags.text_wrap('a\n  \n  \nb'))
    self.assertEqual('a\nb', flags.text_wrap('  a\nb  '))
    self.assertEqual('\na\nb', flags.text_wrap('\na\nb\n'))
    self.assertEqual('\na\nb\n', flags.text_wrap('  \na\nb\n  '))
    self.assertEqual('\na\nb\n', flags.text_wrap('  \na\nb\n\n'))

    # Double newline.
    self.assertEqual('a\n\nb', flags.text_wrap(' a\n\n b'))

    # We respect prefix
    self.assertEqual(' a\n b\n c', flags.text_wrap('a\nb\nc', 80, ' '))
    self.assertEqual('a\n b\n c', flags.text_wrap('a\nb\nc', 80, ' ', ''))

    # tabs
    self.assertEqual('a\n b   c',
                     flags.text_wrap('a\nb\tc', 80, ' ', ''))
    self.assertEqual('a\n bb  c',
                     flags.text_wrap('a\nbb\tc', 80, ' ', ''))
    self.assertEqual('a\n bbb c',
                     flags.text_wrap('a\nbbb\tc', 80, ' ', ''))
    self.assertEqual('a\n bbbb    c',
                     flags.text_wrap('a\nbbbb\tc', 80, ' ', ''))
    self.assertEqual('a\n b\n c\n d',
                     flags.text_wrap('a\nb\tc\td', 3, ' ', ''))
    self.assertEqual('a\n b\n c\n d',
                     flags.text_wrap('a\nb\tc\td', 4, ' ', ''))
    self.assertEqual('a\n b\n c\n d',
                     flags.text_wrap('a\nb\tc\td', 5, ' ', ''))
    self.assertEqual('a\n b   c\n d',
                     flags.text_wrap('a\nb\tc\td', 6, ' ', ''))
    self.assertEqual('a\n b   c\n d',
                     flags.text_wrap('a\nb\tc\td', 7, ' ', ''))
    self.assertEqual('a\n b   c\n d',
                     flags.text_wrap('a\nb\tc\td', 8, ' ', ''))
    self.assertEqual('a\n b   c\n d',
                     flags.text_wrap('a\nb\tc\td', 9, ' ', ''))
    self.assertEqual('a\n b   c   d',
                     flags.text_wrap('a\nb\tc\td', 10, ' ', ''))

    # multiple tabs
    self.assertEqual('a       c',
                     flags.text_wrap('a\t\tc', 80, ' ', ''))

    _helpers._DEFAULT_HELP_WIDTH = default_help_width  # restore

  def test_doc_to_help(self):
    self.assertEqual('', flags.doc_to_help('  '))
    self.assertEqual('', flags.doc_to_help('  \n  '))
    self.assertEqual('a\n\nb', flags.doc_to_help('a\n  \nb'))
    self.assertEqual('a\n\n\nb', flags.doc_to_help('a\n  \n  \nb'))
    self.assertEqual('a b', flags.doc_to_help('  a\nb  '))
    self.assertEqual('a b', flags.doc_to_help('\na\nb\n'))
    self.assertEqual('a\n\nb', flags.doc_to_help('\na\n\nb\n'))
    self.assertEqual('a b', flags.doc_to_help('  \na\nb\n  '))
    # Different first line, one line empty - erm double new line.
    self.assertEqual('a b c\n\nd', flags.doc_to_help('a\n  b\n  c\n\n  d'))
    self.assertEqual('a b\n      c d', flags.doc_to_help('a\n  b\n  \tc\n  d'))
    self.assertEqual('a b\n      c\n      d',
                     flags.doc_to_help('a\n  b\n  \tc\n  \td'))

  def test_doc_to_help_flag_values(self):
    # !!!!!!!!!!!!!!!!!!!!
    # The following doc string is taken as is directly from flags.py:FlagValues
    # The intention of this test is to verify 'live' performance
    # !!!!!!!!!!!!!!!!!!!!
    """Used as a registry for 'Flag' objects.

    A 'FlagValues' can then scan command line arguments, passing flag
    arguments through to the 'Flag' objects that it owns.  It also
    provides easy access to the flag values.  Typically only one
    'FlagValues' object is needed by an application:  flags.FLAGS

    This class is heavily overloaded:

    'Flag' objects are registered via __setitem__:
         FLAGS['longname'] = x   # register a new flag

    The .value member of the registered 'Flag' objects can be accessed as
    members of this 'FlagValues' object, through __getattr__.  Both the
    long and short name of the original 'Flag' objects can be used to
    access its value:
         FLAGS.longname          # parsed flag value
         FLAGS.x                 # parsed flag value (short name)

    Command line arguments are scanned and passed to the registered 'Flag'
    objects through the __call__ method.  Unparsed arguments, including
    argv[0] (e.g. the program name) are returned.
         argv = FLAGS(sys.argv)  # scan command line arguments

    The original registered Flag objects can be retrieved through the use
    """
    doc = flags.doc_to_help(self.test_doc_to_help_flag_values.__doc__)
    # Test the general outline of the converted docs
    lines = doc.splitlines()
    self.assertEqual(17, len(lines))
    empty_lines = [index for index in range(len(lines)) if not lines[index]]
    self.assertEqual([1, 3, 5, 8, 12, 15], empty_lines)
    # test that some starting prefix is kept
    flags_lines = [index for index in range(len(lines))
                   if lines[index].startswith('     FLAGS')]
    self.assertEqual([7, 10, 11], flags_lines)
    # but other, especially common space has been removed
    space_lines = [index for index in range(len(lines))
                   if lines[index] and lines[index][0].isspace()]
    self.assertEqual([7, 10, 11, 14], space_lines)
    # No right space was kept
    rspace_lines = [index for index in range(len(lines))
                    if lines[index] != lines[index].rstrip()]
    self.assertEqual([], rspace_lines)
    # test double spaces are kept
    self.assertEqual(True, lines[2].endswith('application:  flags.FLAGS'))

  def test_text_wrap_raises_on_excessive_indent(self):
    """Ensure an indent longer than line length raises."""
    self.assertRaises(ValueError,
                      flags.text_wrap, 'dummy', length=10, indent=' ' * 10)

  def test_text_wrap_raises_on_excessive_first_line(self):
    """Ensure a first line indent longer than line length raises."""
    self.assertRaises(
        ValueError,
        flags.text_wrap, 'dummy', length=80, firstline_indent=' ' * 80)


if __name__ == '__main__':
  absltest.main()



# Copyright 2017 The Abseil Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for lower/upper bounds validators for numeric flags."""

from unittest import mock
from absl import flags
from absl.flags import _validators
from absl.testing import absltest


class NumericFlagBoundsTest(absltest.TestCase):

  def setUp(self):
    super(NumericFlagBoundsTest, self).setUp()
    self.flag_values = flags.FlagValues()

  def test_no_validator_if_no_bounds(self):
    """Validator is not registered if lower and upper bound are None."""
    with mock.patch.object(_validators, 'register_validator'
                          ) as register_validator:
      flags.DEFINE_integer('positive_flag', None, 'positive int',
                           lower_bound=0, flag_values=self.flag_values)
      register_validator.assert_called_once_with(
          'positive_flag', mock.ANY, flag_values=self.flag_values)
    with mock.patch.object(_validators, 'register_validator'
                          ) as register_validator:
      flags.DEFINE_integer('int_flag', None, 'just int',
                           flag_values=self.flag_values)
      register_validator.assert_not_called()

  def test_success(self):
    flags.DEFINE_integer('int_flag', 5, 'Just integer',
                         flag_values=self.flag_values)
    argv = ('./program', '--int_flag=13')
    self.flag_values(argv)
    self.assertEqual(13, self.flag_values.int_flag)
    self.flag_values.int_flag = 25
    self.assertEqual(25, self.flag_values.int_flag)

  def test_success_if_none(self):
    flags.DEFINE_integer('int_flag', None, '',
                         lower_bound=0, upper_bound=5,
                         flag_values=self.flag_values)
    argv = ('./program',)
    self.flag_values(argv)
    self.assertIsNone(self.flag_values.int_flag)

  def test_success_if_exactly_equals(self):
    flags.DEFINE_float('float_flag', None, '',
                       lower_bound=1, upper_bound=1,
                       flag_values=self.flag_values)
    argv = ('./program', '--float_flag=1')
    self.flag_values(argv)
    self.assertEqual(1, self.flag_values.float_flag)

  def test_exception_if_smaller(self):
    flags.DEFINE_integer('int_flag', None, '',
                         lower_bound=0, upper_bound=5,
                         flag_values=self.flag_values)
    argv = ('./program', '--int_flag=-1')
    try:
      self.flag_values(argv)
    except flags.IllegalFlagValueError as e:
      text = 'flag --int_flag=-1: -1 is not an integer in the range [0, 5]'
      self.assertEqual(text, str(e))


class SettingFlagAfterStartTest(absltest.TestCase):

  def setUp(self):
    self.flag_values = flags.FlagValues()

  def test_success(self):
    flags.DEFINE_integer('int_flag', None, 'Just integer',
                         flag_values=self.flag_values)
    argv = ('./program', '--int_flag=13')
    self.flag_values(argv)
    self.assertEqual(13, self.flag_values.int_flag)
    self.flag_values.int_flag = 25
    self.assertEqual(25, self.flag_values.int_flag)

  def test_exception_if_setting_integer_flag_outside_bounds(self):
    flags.DEFINE_integer('int_flag', None, 'Just integer', lower_bound=0,
                         flag_values=self.flag_values)
    argv = ('./program', '--int_flag=13')
    self.flag_values(argv)
    self.assertEqual(13, self.flag_values.int_flag)
    with self.assertRaises(flags.IllegalFlagValueError):
      self.flag_values.int_flag = -2


if __name__ == '__main__':
  absltest.main()



# Copyright 2017 The Abseil Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Abseil setup configuration."""

import os
import sys

# pylint: disable=g-import-not-at-top
try:
  import setuptools
except ImportError:
  from ez_setup import use_setuptools
  use_setuptools()
  import setuptools
# pylint: enable=g-import-not-at-top

if sys.version_info < (3, 7):
  raise RuntimeError('Python version 3.7+ is required.')

setuptools_version = tuple(
    int(x) for x in setuptools.__version__.split('.')[:2])

additional_kwargs = {}
if setuptools_version >= (24, 2):
  # `python_requires` was added in 24.2, see
  # https://packaging.python.org/guides/distributing-packages-using-setuptools/#python-requires
  additional_kwargs['python_requires'] = '>=3.7'

_README_PATH = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), 'README.md')
with open(_README_PATH, 'rb') as fp:
  LONG_DESCRIPTION = fp.read().decode('utf-8')

setuptools.setup(
    name='absl-py',
    version='2.1.0',
    description=(
        'Abseil Python Common Libraries, '
        'see https://github.com/abseil/abseil-py.'
    ),
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author='The Abseil Authors',
    url='https://github.com/abseil/abseil-py',
    packages=setuptools.find_packages(
        exclude=[
            '*.tests',
            '*.tests.*',
            'tests.*',
            'tests',
        ]
    ),
    include_package_data=True,
    license='Apache 2.0',
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    **additional_kwargs,
)
